package org.openmobileis.oscar;

import java.io.InputStream;
import java.net.URL;

import org.openmobileis.common.util.exception.ServiceException;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.framework.ServiceReference;
import org.osgi.service.packageadmin.PackageAdmin;

public class OscarBundleUtils {
	
	private BundleContext context;
	private String bundlePath = null;
	
	public OscarBundleUtils(BundleContext context, String path) {
		this.context = context;
		this.bundlePath = "file:"+path;
	}
	
	public void setContext(BundleContext context) {
		this.context = context;
	}
	
	public void installWithFileName(String fileName) throws ServiceException {
		if (fileName == null || fileName.trim().length() == 0) {
			throw new ServiceException("OscarBundleManager::installWithFileName : received no filename");
		}
		
		String location = bundlePath+fileName.trim();
		if (!location.endsWith(".jar"));
		location = location+".jar";
		this.install(location);
	}
	
	private void install(String location) throws ServiceException {
		try {
			this.context.installBundle(location);
		} catch (Exception ex) {
			throw new ServiceException(ex);
		}
	}
	
	public boolean isInstalledBundle(String bundleName) throws ServiceException {
		return (this.getBundleIdFromName(bundleName) > 0)?true:false;
	}
	
	public void updateBundleWithName(String bundleName) throws ServiceException {
		this.update(this.getBundleIdFromName(bundleName));
	}
	
	public void updateBundleWithFileName(String bundleFileName) throws ServiceException {
		this.update(this.getBundleIdFromFileName(bundleFileName));
	}
	
	private void update(long bundleId) throws ServiceException {
		this.update(bundleId, null);
	}
	
	public void updateBundleWithName(String bundleName, String location) throws ServiceException {
		this.update(this.getBundleIdFromName(bundleName), location);
	}
	
	public void updateBundleWithFileName(String bundleFileName, String location) throws ServiceException {
		this.update(this.getBundleIdFromFileName(bundleFileName), location);
	}

	private void update(long bundleId, String location) throws ServiceException {
		Bundle bundle = this.context.getBundle(bundleId);
		if (bundle != null) {
			try {
				if (location != null && location.trim().length() > 0) {
                   InputStream is = new URL(location).openStream();
                   bundle.update(is);
				} else {
					bundle.update();
				}
			} catch (Exception ex) {
				throw new ServiceException(ex);
			}
		} else {
			throw new ServiceException("OscarBundleManager::update : Invalid Bundle id : "+bundleId);
		}
	}
	
	public void uninstallBundleWithName(String bundleName) throws ServiceException {
		this.uninstall(this.getBundleIdFromName(bundleName));
	}
	
	public void uninstallBundleWithFileName(String bundleFileName) throws ServiceException {
		this.uninstall(this.getBundleIdFromFileName(bundleFileName));
	}
	
	private void uninstall(long bundleId) throws ServiceException {
		Bundle bundle = this.context.getBundle(bundleId);
		if (bundle != null) {
			try {
				bundle.uninstall();
			} catch (Exception ex) {
				throw new ServiceException(ex);
			}
		} else {
			throw new ServiceException("OscarBundleManager::uninstall : Invalid Bundle id : "+bundleId);
		}
	}
	
	public void startBundleWithName(String bundleName) throws ServiceException {
		this.start(this.getBundleIdFromName(bundleName));
	}
	
	public void startBundleWithFileName(String bundleFileName) throws ServiceException {
		this.start(this.getBundleIdFromFileName(bundleFileName));
	}
	
	private void start(long bundleId) throws ServiceException {
		Bundle bundle = this.context.getBundle(bundleId);
		if (bundle != null) {
			try {
				bundle.start();
			} catch (Exception ex) {
				throw new ServiceException(ex);
			}
		} else {
			throw new ServiceException("OscarBundleManager::start : Invalid Bundle id : "+bundleId);
		}		
	}
	
	public void stopBundleWithName(String bundleName) throws ServiceException {
		this.stop(this.getBundleIdFromName(bundleName));
	}
	
	public void stopBundleWithFileName(String bundleFileName) throws ServiceException {
		this.stop(this.getBundleIdFromFileName(bundleFileName));
	}
	
	private void stop(long bundleId) throws ServiceException {
		Bundle bundle = this.context.getBundle(bundleId);
		if (bundle != null) {
			try {
				bundle.stop();
			} catch (Exception ex) {
				throw new ServiceException(ex);
			}
		} else {
			throw new ServiceException("OscarBundleManager::stop : Invalid Bundle id : "+bundleId);
		}
	}
	
	public void restartBundleWithName(String bundleName) throws ServiceException {
		this.restart(this.getBundleIdFromName(bundleName));
	}
	
	public void restartBundleWithFileName(String bundleFileName) throws ServiceException {
		this.restart(this.getBundleIdFromFileName(bundleFileName));
	}
	
	private void restart(long bundleId) throws ServiceException {
		this.stop(bundleId);
		this.start(bundleId);
	}
	
	public void refreshBundleWithName(String bundleName) throws ServiceException {
		if (bundleName == null) { // We want to refresh all bundles
			this.refresh(-1);
		} else {
			this.refresh(this.getBundleIdFromName(bundleName));
		}
	}
	
	public void refreshBundleWithFileName(String bundleFileName) throws ServiceException {
		if (bundleFileName == null) { // We want to refresh all bundles
			this.refresh(-1);
		} else {
			this.refresh(this.getBundleIdFromFileName(bundleFileName));
		}
	}
	
	private void refresh(long bundleId) throws ServiceException {
        // Get package admin service.
        ServiceReference ref = this.context.getServiceReference(
            org.osgi.service.packageadmin.PackageAdmin.class.getName());
        if (ref == null)
        {
        	throw new ServiceException("OscarBundleManager::refresh : PackageAdmin service is unavailable.");
        }

        PackageAdmin pa = (PackageAdmin) this.context.getService(ref);
        if (pa == null)
        {
        	throw new ServiceException("OscarBundleManager::refresh : PackageAdmin service is unavailable.");
        }

        if (bundleId > 0) { // We want to refresh a specific bundle
        	Bundle bundle = this.context.getBundle(bundleId);
        	if (bundle != null) {
        		pa.refreshPackages(new Bundle[] {bundle});
        	} else {
        		throw new ServiceException("OscarBundleManager::refresh : Bundle not found : "+bundleId);
        	}
        } else {
        	pa.refreshPackages(null);
        }
	}
	
	private long getBundleIdFromName(String bundleName) throws ServiceException {
		
		if (bundleName == null || bundleName.trim().length() == 0) {
			throw new ServiceException("OscarBundleManager::getBundleIdFormName : received empty bundle name");
		}
		
		bundleName = bundleName.trim();
		
		Bundle[] bundles = this.context.getBundles();
		
		if (bundles == null) {
			return -1;
		}
		
		for (int i = 0; i < bundles.length; i++) {
            // Get the bundle name or location.
            String name = (String)
                bundles[i].getHeaders().get(Constants.BUNDLE_NAME);
            
            if (bundleName.equals(name)) {
                return bundles[i].getBundleId();
            }
		}
		
		// No bundle matched the name
		return -1;
	}
	
	private long getBundleIdFromFileName(String bundleFileName) throws ServiceException {
		
		if (bundleFileName == null || bundleFileName.trim().length() == 0) {
			throw new ServiceException("OscarBundleManager::getBundleIdFormName : received empty bundle name");
		}
		
		bundleFileName = bundleFileName.trim();
		
		Bundle[] bundles = this.context.getBundles();
		
		if (bundles == null) {
			return -1;
		}
		
		for (int i = 0; i < bundles.length; i++) {
            // Get the bundle name or location.
            String fullName = (String)
                bundles[i].getLocation();
            
            int slash = fullName.lastIndexOf("/");
            
            String fileNameWithExtension;
            if (slash != -1) {
            	fileNameWithExtension = fullName.substring(slash+1);
            } else {
            	fileNameWithExtension = fullName;
            }
            
            String fileName;
            int dot = fileNameWithExtension.indexOf(".");
            if (dot != -1) {
            	fileName = fileNameWithExtension.substring(0, dot);
            } else {
            	fileName = fileNameWithExtension;
            }
            
            if (bundleFileName.equals(fileName)) {
                return bundles[i].getBundleId();
            }
		}
		
		// No bundle matched the name
		return -1;
	}
	
	public void shutdown() throws ServiceException {
		Bundle bundle = this.context.getBundle(0);
		try {
			bundle.stop();
		} catch (Exception ex) {
			throw new ServiceException(ex);
		}
	}
	
}
