package org.openmobileis.oscar;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Properties;

import org.openmobileis.common.util.PropertiesManager;
import org.ungoverned.oscar.Noscar;
import org.ungoverned.oscar.cache.DefaultBundleCache;
import org.ungoverned.oscar.util.CaseInsensitiveMap;
import org.ungoverned.oscar.util.MutablePropertyResolverImpl;

public class StartOscar {
	
    public static final String CONFIG_PROPERTIES_FILE_VALUE = "oscar.properties";


	public static void main(String[] args) {
		
		System.out.println("user.dir : "+System.getProperty("user.dir"));
		Properties configProps = null;
		
        // Read configuration properties.
		if (args != null && args.length > 0) {
			configProps = StartOscar.readConfigProperties(args[0]);
		} else {
			configProps = StartOscar.readConfigProperties();
		}
		
        // See if the profile name property was specified.
        String profileName = configProps.getProperty(DefaultBundleCache.CACHE_PROFILE_PROP);

        // See if the profile directory property was specified.
        String profileDirName = configProps.getProperty(DefaultBundleCache.CACHE_PROFILE_DIR_PROP);

        // Print welcome banner.
        System.out.println("\nWelcome to ObjectWeb Open Mobile IS Oscar.");
        System.out.println("==========================================\n");

        // If no profile or profile directory is specified in the
        // properties, then ask for a profile name.
        if ((profileName == null) && (profileDirName == null))
        {
            System.out.print("Enter profile name: ");
            BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
            try
            {
                profileName = in.readLine();
            }
            catch (IOException ex)
            {
                System.err.println("Could not read input.");
                System.exit(-1);
            }
            System.out.println("");
            if (profileName.length() != 0)
            {
                configProps.setProperty(DefaultBundleCache.CACHE_PROFILE_PROP, profileName);
            }
        }

        // A profile directory or name must be specified.
        if ((profileDirName == null) && (profileName.length() == 0))
        {
            System.err.println("You must specify a profile name or directory.");
            System.exit(-1);
        }

        try
        {
            // Now create an instance of Oscar.
            Noscar oscar = new Noscar();
            oscar.start(
                new MutablePropertyResolverImpl(new CaseInsensitiveMap(configProps)),
                null);
        }
        catch (Exception ex)
        {
            System.err.println("Could not create Oscar: " + ex);
            ex.printStackTrace();
            System.exit(-1);
        }
	}
	
	public static Properties readConfigProperties() {
		return readConfigProperties(null);
	}
	
    public static Properties readConfigProperties(String userdir)
    {
        // See if the property URL was specified as a property.
        URL propURL = null;
        String propFileName = null;
        try
        {
        	if (userdir != null && userdir.trim().length() > 0) {
        		System.setProperty("user.dir", userdir.trim());
        	}
        	propURL = new File(System.getProperty("user.dir") + File.separator + "WEB-INF" + File.separator + "conf" + File.separator + "properties" + File.separator + CONFIG_PROPERTIES_FILE_VALUE).toURL();
        	propFileName = System.getProperty("user.dir") + File.separator + "WEB-INF" + File.separator + "conf" + File.separator + "properties" + File.separator + CONFIG_PROPERTIES_FILE_VALUE;
        	System.out.println("propFileName : "+propFileName);
        }
        catch (MalformedURLException ex)
        {
        	System.err.print("StartOscar: " + ex);
        	return null;
        }

        // Read the properties file.
        Properties props = new Properties();
        InputStream is = null;
        try
        {
        	// HERE, we use our propertiesManager instead of oscar's method to retrieve
        	// properties, as we want to take advantage of the ${user.dir} variable.
            //is = propURL.openConnection().getInputStream();
            props = PropertiesManager.getManager().getPropertiesFromPath(propFileName);
            //props.load(is);
            //is.close();
        }
        catch (FileNotFoundException ex)
        {
            // Ignore file not found.
        }
        catch (Exception ex)
        {
            System.err.println(
                "StartOscar: Error loading system properties from " + propURL);
            System.err.println("StartOscar: " + ex);
            try
            {
                if (is != null) is.close();
            }
            catch (IOException ex2)
            {
                // Nothing we can do.
            }
            return null;
        }

        // Perform variable substitution for system properties.
        for (Enumeration e = props.propertyNames(); e.hasMoreElements(); )
        {
            String name = (String) e.nextElement();
            props.setProperty(name, substVars((String) props.getProperty(name)));
        }

        return props;
    }
    
    private static final String DELIM_START = "${";
    private static final char DELIM_STOP  = '}';
    private static final int DELIM_START_LEN = 2;
    private static final int DELIM_STOP_LEN  = 1;
    
    public static String substVars(String val)
    throws IllegalArgumentException
{
    StringBuffer sbuf = new StringBuffer();

    if (val == null)
    {
        return val;
    }

    int i = 0;
    int j, k;

    while (true)
    {
        j = val.indexOf(DELIM_START, i);
        if (j == -1)
        {
            if (i == 0)
            {
                return val;
            }
            else
            {
                sbuf.append(val.substring(i, val.length()));
                return sbuf.toString();
            }
        }
        else
        {
            sbuf.append(val.substring(i, j));
            k = val.indexOf(DELIM_STOP, j);
            if (k == -1)
            {
                throw new IllegalArgumentException(
                '"' + val +
                "\" has no closing brace. Opening brace at position "
                + j + '.');
            }
            else
            {
                j += DELIM_START_LEN;
                String key = val.substring(j, k);
                // Try system properties.
                String replacement = System.getProperty(key, null);
                if (replacement != null)
                {
                    sbuf.append(replacement);
                }
                i = k + DELIM_STOP_LEN;
            }
        }
    }
}

}
