/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2005 Creation P.Delrieu
 * 
 */


package org.openmobileis.common.intl;

import java.util.Properties;
import java.util.TimeZone;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;

import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.embedded.util.ServicePropertiesManager;

import java.util.Locale;


/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Manage international resource.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class IntlResourceManager {
  /**
   * default extension extension : nothing
   */
  public final static String DEFAULT="fr"; 
  /**
   * France local extension :fr
   */
  public final static String FR="fr"; 
  /**
   * United State local extension :us
   */
  public final static String US="us";
  /**
   * England local extension :en
   */
  public final static String EN="en";
  /**
   * Belgium local extension :be
   */
  public final static String BE="be";
  /**
   * Switzerland local extension :sw
   */
  public final static String SW="sw";
  /**
   * Deutchland local extension :de
   */
  public final static String DE="de";
  /**
   * Italian local extension :it
   */
  public final static String IT="it";
  /**
   * Espana local extension :es
   */
  public final static String ES="es";

  private static IntlResourceManager manager;
  
  private Properties localizedProperties;
  
  private String systemLocal;
  
  /** Creates a new instance of IntlResourceManager */
  private IntlResourceManager() {
    localizedProperties = new Properties();
    systemLocal = ServicePropertiesManager.getManager().getProperty("intlmanager", "local");
    if (systemLocal ==null) {
    	systemLocal = PropertiesManager.getManager().getProperty("serve.intl.defaultinitlocal", IntlResourceManager.DEFAULT);
    }
		//set VM parameters
		Locale.setDefault(this.getManagerLocale());
		TimeZone.setDefault(TimeZone.getTimeZone("ECT"));
  }
  
  public static IntlResourceManager getManager()  {
    if (manager == null)  {
      synchronized(IntlResourceManager.class) {
        if (manager == null)  {
          manager = new IntlResourceManager();
					ApplicationContextManager.getManager().addManager(manager);
        }
      }
    }
    return manager;
  }
  
  /**
   * load the properties file with specified locale. Localized filename are of type filename.localExtension.
   * If the localised file is not found try to load the filename (without extension). If it's failed return java.io.FileNotFoundException.
   * @param String fileName : non localized filename
   * @param local : locale of the file to load
   * @ return the loaded properties
   * @exception : java.io.FileNotFoundException if the localized file and fileName is not found.
   * @exception : java.io.IOException if an error occurs.
   *
   */
  public Properties loadLocalizedProperties(String fileName, String local)  throws java.io.FileNotFoundException, java.io.IOException {
    Properties props = new Properties();
    this.loadLocalizedPropertiesInProps(props, fileName, local);
    return props;
  }
  
  private void loadLocalizedPropertiesInProps(Properties props, String fileName, String local)  throws java.io.FileNotFoundException, java.io.IOException {
    File file = new File(fileName+"."+local.toLowerCase());
    if (!file.exists()) {
      file = new File(fileName); 
    }
    	InputStream in = new BufferedInputStream(new FileInputStream(file));
    try	{
    	props.load(in);
    } finally	{
    	in.close();
    }
  }
  
  /**
   * return the fileName transformed with the system local. (ie:fileName.systemlocal).
   * if the localized file does not exist return the specified filename.
   */
  public String getLocalizedFileName(String fileName){
    String localized = fileName+"."+systemLocal;
    File file = new File(localized);
    if (!file.exists()) {
      localized = fileName;
    }
    return localized;

  }
  
  /**
   * return the fileName without the current locale extension if present.
   * if not return the file name.
   */
  public String removeLocaleFromFileName(String fileName){
    if (fileName.endsWith(systemLocal))  {
      int index = fileName.lastIndexOf( '.' );
      if (index != -1)  {
        fileName = fileName.substring(0, index);
      }
    }
    return fileName;

  }
  /**
   * load the properties file with specified locale in the IntlResourceManager properties . Localized filename are of type filename.localExtension.
   * If the localised file is not found try to load the filename (without extension). If it's failed return java.io.FileNotFoundException.
   * @param String fileName : non localized filename
   * @param local : locale of the file to load
   * @ return the loaded properties
   * @exception : java.io.FileNotFoundException if the localized file and fileName is not found.
   * @exception : java.io.IOException if an error occurs.
   *
   */
  public void addLocalizedProperties(String propertiesFile, String local) throws java.io.FileNotFoundException, java.io.IOException{
    this.loadLocalizedPropertiesInProps(localizedProperties, propertiesFile, local);
  }
  
  public String getLocalizedProperty(String propKey)  {
    return localizedProperties.getProperty(propKey);
  }
  
  public void setLocalizedProperty(String key, String value)  {
    localizedProperties.put(key, value);
  }
  
  public String getLocalizedProperty(String propKey, String defaultValue)  {
    String val = localizedProperties.getProperty(propKey);
    if (val == null)  {
      val = defaultValue;
    }
    return val;
  }
  
   /**
    *return the local defined in the webserver property server.locale. If not defined default is use.
    *
    **/
  public String getManagerLocaleString()  {
    return systemLocal;
  }
  
  public void setManagerLocal(String local)	{
  	systemLocal=local;
		ServicePropertiesManager.getManager().saveProperty("intlmanager", "local", systemLocal);
		//set VM parameters
		Locale.setDefault(this.getManagerLocale());
  }
  
  public Locale getManagerLocale()  {
    if (systemLocal.equals(FR)) {
      return Locale.FRANCE;
    } else if (systemLocal.equals(EN))  {
      return Locale.UK;
    } else if (systemLocal.equals(US))  {
      return Locale.US;
    } else if (systemLocal.equals(BE))  {
      return Locale.FRANCE;
    } else if (systemLocal.equals(IT))  {
      return Locale.ITALY;
    } else if (systemLocal.equals(DE))  {
      return Locale.GERMANY;
    }
    return Locale.US;
  }
 
  
}
