/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */


package org.openmobileis.common.util;

import org.openmobileis.common.util.collection.Array;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Implement a generic object pool. The pool contain a set of object and client get and dispose object depending on there need. When  an object is retreive fro; the pool (getObject ()), it must be disposed. If not it can leads to memory link ( object not garbage collected). So to insure this behaviour put object use in a try / finally clause.
 * example :
 * <p><tt>
 Object obj = ObjectPool.getObject();
 try {
   // use the object
 }
 finally [
   ObjectPool.disposeObject(obj);
 }
 </tt></p>
 *
 * A pool can only create a single type of object. The size of the pool is initialized with the minimum size at the begining and grow by the specified increment size.
 *
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public abstract class ObjectPool {
  protected java.util.Stack pool_;
  protected Array arrayPool;
  protected Class objectClass_;
  protected int minSize_, incSize_;

  /**
   * @param objectClass
   * @param minSize
   * @param incSize
   * @return
   * @exception
   * @author
   * @see
   * @since
   * @version
   * @roseuid 38C5242C0241
   */
  protected ObjectPool(String objectClass, int minSize, int incSize) throws ClassNotFoundException, IllegalAccessException, InstantiationException  {
    pool_ = new java.util.Stack();
      objectClass_ = Class.forName(objectClass);
    if (minSize == 0) {
      minSize_ = 10;
    } else  {
      minSize_ = minSize;
    }
    incSize_ = incSize;
    arrayPool = new Array(minSize);
    createInPool(minSize);
  }

  private final void createInPool(int size) throws IllegalAccessException, InstantiationException {
    for (int i=0; i<size; i++)  {
      Object obj = createNewInstance(objectClass_);
      pool_.push(obj);
      arrayPool.add(obj);
    }
  }

  /**
  	* A factory method that you may override to instantiate the proper object
  	* Default implementation does a newInstance()
  */
  protected Object createNewInstance(Class objectClass) throws IllegalAccessException, InstantiationException {
  	return objectClass.newInstance();
  }

  /**
   * @return Object
   * @exception
   * @author
   * @see
   * @since
   * @version
   * Return an available object from the pool. If no object is available create a new one depending on the increment argument. If the pool can not grow (increment = 0) return null.
   * @roseuid 38C5224B0147
   */
  protected final synchronized Object getObject() {
    try {
      return pool_.pop();
    } catch(java.util.EmptyStackException ex) {
      try {
        createInPool(minSize_);
		    return pool_.pop();
      } catch (Exception e) {
        // already catch in constructor call
      }
    }
    return null;
  }

  /**
   * @param object
   * @return void
   * @exception
   * @author
   * @see
   * @since
   * @version
   * dispose the object. The object return to the pool and become available.
   * @roseuid 38C5229D001E
   */
  public synchronized void disposeObject(Object object) {
    if (object != null) {
      pool_.push( object );
    }
  }

}