/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */

package org.openmobileis.common.util.collection;

import org.openmobileis.common.util.OpenMISSerializable;

/**
 * Use to store an Array of objects.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */

public class Array implements	FilterCollection,	OpenMISSerializable {
	protected static final long serialVersionUID = 5521257935120563452L;

	/**
	 * The array buffer into which the elements of the ArrayList are stored.
	 * The capacity of the ArrayList is the length of this array buffer.
	 */
	protected Object elementData[];

	/**
	 * The size of the ArrayList (the number of elements it contains).
	 */
	protected int size;

	private Array(Array array) {
		Object[] retElement = new Object[array.elementData.length];
		System.arraycopy(array.elementData, 0, retElement, 0,
				array.elementData.length);
		elementData = retElement;
		size = array.size;
	}

	public Array(Object[] array) {
		elementData = array;
		size = array.length;
	}

	/**
	 * Constructs an empty ArrayList with the specified initial capacity.
	 *
	 * @param   initialCapacity   the initial capacity of the ArrayList.
	 * @exception IllegalArgumentException if the specified initial capacity
	 *            is negative
	 */
	public Array(int initialCapacity) {
		super();
		if (initialCapacity < 0)
			throw new IllegalArgumentException("Illegal Capacity: "
					+ initialCapacity);
		this.elementData = new Object[initialCapacity];
	}

	/**
	 * Constructs an empty ArrayList.Initial capacity is 10
	 */
	public Array() {
		this(10);
	}

	public void clear() {
		this.elementData = new Object[this.elementData.length];
		size = 0;
	}

	protected Object[] getArrayElements() {
		return elementData;
	}

	protected void setArrayElements(Object[] array) {
		elementData = array;
	}

	/**
	 * Increases the capacity of this <tt>ArrayList</tt> instance, if
	 * necessary, to ensure  that it can hold at least the number of elements
	 * specified by the minimum capacity argument.
	 *
	 * @param   minCapacity   the desired minimum capacity.
	 */
	public void ensureCapacity(int minCapacity) {
		int oldCapacity = elementData.length;
		if (minCapacity >= oldCapacity) {
			Object oldData[] = elementData;
			int newCapacity = (oldCapacity * 3) / 2 + 1;
			if (newCapacity < minCapacity)
				newCapacity = minCapacity;
			elementData = new Object[newCapacity];
			System.arraycopy(oldData, 0, elementData, 0, size);
		}
	}

	/**
	 * Returns the number of elements in this ArrayList.
	 *
	 * @return  the number of elements in this ArrayList.
	 */
	public int size() {
		return size;
	}

	/**
	 * set the size of the array. All element after the size are removed.
	 */
	public void setSize(int size) {
		this.size = size;
	}

	/**
	 * Tests if this ArrayList has no elements.
	 *
	 * @return  <tt>true</tt> if this ArrayList has no elements;
	 *          <tt>false</tt> otherwise.
	 */
	public boolean isEmpty() {
		return size == 0;
	}

	/**
	 * Returns an array containing all of the elements in this ArrayList
	 * in the correct order.
	 *
	 * @return an array containing all of the elements in this ArrayList
	 * 	       in the correct order.
	 */
	public Object[] toArray() {
		Object[] result = new Object[size];
		System.arraycopy(elementData, 0, result, 0, size);
		return result;
	}

	/**
	 * Returns an array containing all of the elements in this ArrayList in the
	 * correct order.  The runtime type of the returned array is that of the
	 * specified array.  If the ArrayList fits in the specified array, it is
	 * returned therein.  Otherwise, a new array is allocated with the runtime
	 * type of the specified array and the size of this ArrayList.<p>
	 *
	 * If the ArrayList fits in the specified array with room to spare (i.e., the
	 * ArrayList has more elements than the array), the element in the array
	 * immediately following the end of the collection is set to
	 * <tt>null</tt>.
	 *
	 * @param a the array into which the elements of the ArrayList are to
	 *		be stored, if it is big enough; otherwise, a new array of the
	 * 		same runtime type is allocated for this purpose.
	 * @return an array containing the elements of the ArrayList.
	 * @throws ArrayStoreException if the runtime type of a is not a supertype
	 *         of the runtime type of every element in this ArrayList.
	 */
	public Object[] toArray(Object a[]) {
		if (a.length < size)
			a = (Object[]) java.lang.reflect.Array.newInstance(a.getClass()
					.getComponentType(), size);

		System.arraycopy(elementData, 0, a, 0, size);

		if (a.length > size)
			a[size] = null;

		return a;
	}

	// Positional Access Operations

	/**
	 * Returns the element at the specified position in this ArrayList.
	 *
	 * @param  index index of element to return.
	 * @return the element at the specified position in this ArrayList.
	 * @throws    IndexOutOfBoundsException if index is out of range <tt>(index
	 * 		  &lt; 0 || index &gt;= size())</tt>.
	 */
	public Object get(int index) {
		RangeCheck(index);

		return elementData[index];
	}

	/**
	 * Appends the specified element to the end of this ArrayList.
	 *
	 * @param o element to be appended to this ArrayList.
	 * @return <tt>true</tt>.
	 */
	public boolean add(Object o) {
		ensureCapacity(size + 1); // Increments modCount!!
		elementData[size++] = o;
		return true;
	}

	/**
	 * replace at the specified index the element.
	 * if there is no element at index add the new one at the end of the array.
	 *
	 * @param index index in the array where element is replaced.
	 * @param o element to be replaced to this ArrayList.
	 * @return <tt>true</tt>.
	 */
	public boolean replace(int index, Object o) {
		if (index >= size) {
			return this.add(o);
		}
		elementData[index] = o;
		return true;
	}

	/**
	 * Appends the specified element to the end of this ArrayList.
	 *
	 * @param o element to be appended to this ArrayList.
	 * @return <tt>true</tt>.
	 */
	public boolean add(int index, Object o) {
		ensureCapacity(index); // Increments modCount!!
		elementData[index] = o;
		if (index >= size)
			size = index + 1;
		return true;
	}

	/**
	 * return the last added object in the array.
	 *
	 * @param 
	 * @return last added Object.
	 */
	public Object getLastAdded() {
		if (size > 0) {
			return elementData[size - 1];
		}
		return null;
	}

	/**
	 * return true if the array contains the specified object false otherwise.
	 *
	 * @param Object o : the object to be tested. O must implements equals method
	 * @return <tt>true</tt> if the specified object is inside the array.
	 */
	public boolean contains(Object o) {
		for (int i = 0; i < size; i++) {
			if (elementData[i].equals(o)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Searches for the first occurence of the given argument, testing for equality using the equals method.
	 * return the index of the occurence, -1 if not found
	 *
	 * @param Object o : the object to be searched. O must implements equals method
	 * @return index of the first occurence, -1 if not found.
	 */
	public int indexOf(Object o) {
		for (int i = 0; i < size; i++) {
			if (elementData[i].equals(o)) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Appends the specified ArrayList to the end of this ArrayList.
	 *
	 * @param a ArrayList to be appended to this ArrayList..
	 * @return <tt>true</tt>.
	 */
	public boolean addArray(Array a) {
		int newSize = size + a.size;
		Object[] newArray = new Object[newSize];
		System.arraycopy(elementData, 0, newArray, 0, size);
		System.arraycopy(a.elementData, 0, newArray, size, a.size);
		this.size = newSize;
		this.elementData = newArray;
		return true;
	}

	/**
	 * Removes the element at the specified position in this ArrayList.
	 * Shifts any subsequent elements to the left (subtracts one from their
	 * indices).
	 *
	 * @param index the index of the element to removed.
	 * @return the element that was removed from the ArrayList.
	 * @throws    IndexOutOfBoundsException if index out of range <tt>(index
	 * 		  &lt; 0 || index &gt;= size())</tt>.
	 */
	public Object remove(int index) {
		RangeCheck(index);
		Object oldValue = elementData[index];

		int numMoved = size - index - 1;
		if (numMoved > 0)
			System.arraycopy(elementData, index + 1, elementData, index,
					numMoved);
		elementData[--size] = null; // Let gc do its work

		return oldValue;
	}

	/**
	 * Removes the element in this ArrayList.
	 * To find the element, array element is compared with equals method until found.
	 *
	 * @param  element to removed.
	 * @return the element that was removed from the ArrayList. Null if not found.
	 */
	public Object remove(Object element) {
		Object oldValue = null;
		for (int i = 0; i < elementData.length; i++) {
			if (elementData[i] == null)
				break;
			if (elementData[i].equals(element)) {
				oldValue = remove(i);
				break;
			}
		}
		return oldValue;
	}

	public void removeAll() {
		size = 0;
	}

	/**
	 * Check if the given index is in range.  If not, throw an
	 * IndexOutOfBoundsException exception.
	 */
	private void RangeCheck(int index) {
		if (index >= size || index < 0)
			throw new IndexOutOfBoundsException("Index: " + index + ", Size: "
					+ size);
	}

	// tests
	public static void main(String[] args) {
		Array array = new Array(2);
		array.add("Chaine 1");
		array.add("chaine2");
		array.add("chaine3");
		array.add(2, "Chaine 4");
		System.out.println("Value at 2 " + array.get(2));
		System.out.println("Value at 3 " + array.get(3));
		array.add(2, "chaine2bis");
		System.out.println("Value at 2 " + array.get(2));
		array.remove(0);
		for (int i = 0; i < array.size; i++) {
			System.out.println("Value at " + i + " : " + array.get(i));
		}

		String[] temp = (String[]) array.toArray(new String[array.size]);
		for (int i = 0; i < temp.length; i++) {
			System.out.println("Value at " + i + " : " + temp[i]);
		}
	}

	public Object clone() {
		return new Array(this);

	}

	public boolean equals(Object obj) {
		if (obj == this)
			return true;

		if (!(obj instanceof Array))
			return false;
		Array t = (Array) obj;
		if (t.size() != size)
			return false;

		Object[] arrayObj = t.getArrayElements();

		for (int i = 0; i < size; i++) {
			if (elementData[i] == null) {
				if (arrayObj[i] != null) {
					return false;
				}
			} else {
				if (!elementData[i].equals(arrayObj[i])) {
					return false;
				}
			}
		}

		return true;
	}

}