/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */


package org.openmobileis.common.util.collection;

import org.openmobileis.common.util.OpenMISSerializable;

/**
 * Use to store an Array of long.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */

public final class LongArray implements OpenMISSerializable {
	protected static final long serialVersionUID = 5521257935120563452L;

	/**
	 * The array buffer into which the elements of the ArrayList are stored.
	 * The capacity of the ArrayList is the length of this array buffer.
	 */
	private long elementData[];

	/**
	 * The size of the Array (the number of elements it contains).
	 */
	private int size;

	/**
	 * Constructs an empty list with the specified initial capacity.
	 *
	 * @param   initialCapacity   the initial capacity of the list.
	 * @exception IllegalArgumentException if the specified initial capacity
	 *            is negative
	 */
	public LongArray(int initialCapacity) {
		super();
		if (initialCapacity < 0)
			throw new IllegalArgumentException("Illegal Capacity: "
					+ initialCapacity);
		this.elementData = new long[initialCapacity];
	}

	/**
	 * Constructs an empty list.
	 */
	public LongArray() {
		this(10);
	}

	public void clear() {
		this.elementData = new long[this.elementData.length];
		size = 0;
	}

	protected long[] getArrayElements() {
		return elementData;
	}

	protected void setArrayElements(long[] array) {
		elementData = array;
	}

	/**
	 * Increases the capacity of this <tt>ArrayList</tt> instance, if
	 * necessary, to ensure  that it can hold at least the number of elements
	 * specified by the minimum capacity argument.
	 *
	 * @param   minCapacity   the desired minimum capacity.
	 */
	public void ensureCapacity(int minCapacity) {
		int oldCapacity = elementData.length;
		if (minCapacity > oldCapacity) {
			long oldData[] = elementData;
			int newCapacity = (oldCapacity * 3) / 2 + 1;
			if (newCapacity < minCapacity)
				newCapacity = minCapacity;
			elementData = new long[newCapacity];
			System.arraycopy(oldData, 0, elementData, 0, size);
		}
	}

	/**
	 * Returns the number of elements in this list.
	 *
	 * @return  the number of elements in this list.
	 */
	public int size() {
		return size;
	}

	/**
	 * Tests if this list has no elements.
	 *
	 * @return  <tt>true</tt> if this list has no elements;
	 *          <tt>false</tt> otherwise.
	 */
	public boolean isEmpty() {
		return size == 0;
	}

	/**
	 * Returns an array containing all of the elements in this list
	 * in the correct order.
	 *
	 * @return an array containing all of the elements in this list
	 * 	       in the correct order.
	 */
	public long[] toArray() {
		long[] result = new long[size];
		System.arraycopy(elementData, 0, result, 0, size);
		return result;
	}

	// Positional Access Operations

	/**
	 * Returns the element at the specified position in this list.
	 *
	 * @param  index index of element to return.
	 * @return the element at the specified position in this list.
	 * @throws    IndexOutOfBoundsException if index is out of range <tt>(index
	 * 		  &lt; 0 || index &gt;= size())</tt>.
	 */
	public long get(int index) {
		RangeCheck(index);

		return elementData[index];
	}

	/**
	 * Appends the specified element to the end of this list.
	 *
	 * @param o element to be appended to this list.
	 * @return <tt>true</tt> (as per the general contract of Collection.add).
	 */
	public boolean add(long o) {
		ensureCapacity(size + 1); // Increments modCount!!
		elementData[size++] = o;
		return true;
	}

	/**
	 * Appends the specified LongArray to the end of this list.
	 *
	 * @param array long to add.
	 * @return <tt>true</tt> (as per the general contract of Collection.add).
	 */
	public boolean add(LongArray array) {
		ensureCapacity(size + array.size()); // Increments modCount!!
		System.arraycopy(array.getArrayElements(), 0, elementData, size, array
				.size());
		size += array.size();
		return true;
	}

	/**
	 * Appends the specified LongArray to the end of this list.
	 *
	 * @param array long to add.
	 * @return <tt>true</tt> (as per the general contract of Collection.add).
	 */
	public boolean add(long[] array) {
		ensureCapacity(size + array.length); // Increments modCount!!
		System.arraycopy(array, 0, elementData, size, array.length);
		size += array.length;
		return true;
	}

	/**
	 * replace at the specified index the element.
	 * if there is no element at index add the new one at the end of the array.
	 *
	 * @param index index in the array where element is replaced.
	 * @param o element to be replaced to this ArrayList.
	 * @return <tt>true</tt>.
	 */
	public boolean replace(int index, long o) {
		if (index >= size) {
			return this.add(o);
		}
		elementData[index] = o;
		return true;
	}

	/**
	 * Appends the specified element to the end of this list.
	 *
	 * @param o element to be appended to this list.
	 * @return <tt>true</tt> (as per the general contract of Collection.add).
	 */
	public boolean add(int index, long o) {
		ensureCapacity(index); // Increments modCount!!
		elementData[index] = o;
		if (index >= size)
			size = index + 1;
		return true;
	}

	/**
	 * Removes the element at the specified position in this list.
	 * Shifts any subsequent elements to the left (subtracts one from their
	 * indices).
	 *
	 * @param index the index of the element to removed.
	 * @return the element that was removed from the list.
	 * @throws    IndexOutOfBoundsException if index out of range <tt>(index
	 * 		  &lt; 0 || index &gt;= size())</tt>.
	 */
	public long remove(int index) {
		RangeCheck(index);
		long oldValue = elementData[index];

		int numMoved = size - index - 1;
		if (numMoved > 0)
			System.arraycopy(elementData, index + 1, elementData, index,
					numMoved);
		elementData[--size] = 0; // Let gc do its work

		return oldValue;
	}

	public void addIntersect(long[] array) {
		long[] thisarray = this.toArray();
		long[] newArray = LongArray.intersectLongArray(thisarray, array);
		this.elementData = newArray;
		this.size = this.elementData.length;
	}

	public void addUnion(long[] array) {
		long[] thisarray = this.toArray();
		long[] newArray = LongArray.unionLongArray(thisarray, array);
		this.elementData = newArray;
		this.size = this.elementData.length;
	}

	/**
	 * Check if the given index is in range.  If not, throw an appropriate
	 * runtime exception.
	 */
	private void RangeCheck(int index) {
		if (index >= size || index < 0)
			throw new IndexOutOfBoundsException("Index: " + index + ", Size: "
					+ size);
	}

	public static void shellsort(long[] data) {
		// sort using Shellsort
		int h = 1;
		int i = 0;
		int n = data.length;

		while (h <= n)
			h = h * 3 + 1;
		while (h > 1) {
			h /= 3;
			long v;
			for (i = h + 1; i <= n; i++) {
				v = data[i - 1];
				int j = i;
				while ((data[j - h - 1] - v) > 0) {
					data[j - 1] = data[j - h - 1];
					j -= h;
					if (j <= h)
						break;
				}
				data[j - 1] = v;
			}
		}
	}

	public static long[] intersectLongArray(long[] array1, long[] array2) {
		if (array1.length == 0)
			return array1;
		if (array2.length == 0)
			return array2;
		if (array1.length > 1)
			LongArray.shellsort(array1);
		if (array2.length > 1)
			LongArray.shellsort(array2);

		LongArray retArray = new LongArray(array1.length);
		long ret = 0;
		int i = 0, j = 0;
		while ((i < array1.length) && (j < array2.length)) {
			ret = array1[i] - array2[j];
			if (ret == 0) {
				retArray.add(array1[i]);
				i++;
				j++;
			} else if (ret > 0) {
				j++;
			} else {
				i++;
			}
		}
		return retArray.toArray();
	}

	public static long[] unionLongArray(long[] array1, long[] array2) {
		if (array1.length == 0)
			return array2;
		if (array2.length == 0)
			return array1;
		if (array1.length > 1)
			LongArray.shellsort(array1);
		if (array2.length > 1)
			LongArray.shellsort(array2);
		LongArray retArray = new LongArray(array1.length + array2.length);
		long ret = 0;
		int i = 0, j = 0;
		while ((i < array1.length) && (j < array2.length)) {
			ret = array1[i] - array2[j];
			if (ret == 0) {
				retArray.add(array1[i]);
				i++;
				j++;
			} else if (ret > 0) {
				retArray.add(array2[j]);
				j++;
			} else {
				retArray.add(array1[i]);
				i++;
			}
		}
		if (i < array1.length) {
			long[] newArray = new long[array1.length - i];
			System.arraycopy(array1, i, newArray, 0, newArray.length);
			retArray.add(newArray);
		}
		if (j < array2.length) {
			long[] newArray = new long[array2.length - j];
			System.arraycopy(array2, j, newArray, 0, newArray.length);
			retArray.add(newArray);
		}

		return retArray.toArray();
	}

	/**
	 * return a array containing all element of the second array that is not in the first one.
	 * @param array1 first array
	 * @param array2 second array
	 * @return array of long that contained in the second array and not in the first.
	 */
	public static long[] diffLongArray(long[] array1, long[] array2) {
		if (array1.length == 0)
			return array2;
		if (array2.length == 0)
			return array2;
		if (array1.length > 1)
			LongArray.shellsort(array1);
		if (array2.length > 1)
			LongArray.shellsort(array2);

		LongArray retArray = new LongArray(array2.length);
		long ret = 0;
		int i = 0, j = 0;
		while ((i < array1.length) && (j < array2.length)) {
			ret = array1[i] - array2[j];
			if (ret == 0) {
				i++;
				j++;
			} else if (ret > 0) {
				retArray.add(array2[j]);
				j++;
			} else {
				i++;
			}
		}
		// add end of second array
		for (i = j; i < array2.length; i++) {
			retArray.add(array2[i]);
		}
		return retArray.toArray();
	}
}