/*
 * $Id: Constraint.java,v 1.1 2005/07/19 12:27:49 dev Exp $
 * S.O.D.A.  -  Simple Object Database Access
 */
package  org.odbms;


/**
 * constraint to limit the objects returned upon
 * Query#execute() query execution.
 * <br><br>
 * Constraints are constructed by calling 
 * Query#constrain Query.constrain().
 * <br><br>
 * Constraints can be joined with the methods #and and()
 * and #or or().
 * <br><br>
 * The methods to modify the constraint evaluation algorithm may
 * be merged, to construct combined evaluation rules.
 * Examples:
 * <ul>
 *   <li> <code>Constraint#smaller().equal()</code> for "smaller or equal" </li>
 *   <li> <code>Constraint#not().like()</code> for "not like" </li>
 *   <li> <code>Constraint#not().greater().equal()</code> for "not greater or equal" </li>
 * </ul>
 */
public interface Constraint {

    /**
	 * links two Constraints for AND evaluation.
     * @param with the other Constraint
     * @return a new Constraint, that can be used for further calls
	 * to and() and or()}
     */
    public Constraint and (Constraint with);


    /**
	 * links two Constraints for OR evaluation.
     * @param with the other Constraint
     * @return a new Constraint, that can be used for further calls
     * to  and() and  or()
     */
    public Constraint or (Constraint with);


    /**
     * sets the evaluation mode to <code>==</code>.
     * @return this Constraint to allow the chaining of method calls.
     */
    public Constraint equal ();


    /**
     * sets the evaluation mode to <code>&gt;</code>.
     * @return this Constraint to allow the chaining of method calls.
     */
    public Constraint greater ();

    /**
     * sets the evaluation mode to <code>&lt;</code>.
     * @return this Constraint to allow the chaining of method calls.
     */
    public Constraint smaller ();


    /**
     * sets the evaluation mode to identity comparison.
     * @return this Constraint to allow the chaining of method calls.
     */
    public Constraint identity ();
	
	
    /**
     * sets the evaluation mode to "like" comparison.
     * @return this Constraint to allow the chaining of method calls.
     */
    public Constraint like ();
	
	
    /**
     * sets the evaluation mode to containment comparison.
	 * <br><br>
	 * Evaluation is dependant on the constrained query node:
	 * <dl>
	 *  <dt><code>String</code></dt>
	 *   <dd>the persistent object is tested to contain a substring.</dd>
	 *  <dt>arrays, java.util.Collection collections</dt>
	 *   <dd>the persistent object is tested to contain all elements of
	 *      the constraining object.</dt>
	 * </dl>
     * @return this Constraint to allow the chaining of method calls.
     */
    public Constraint contains ();


    /**
     * turns on not() comparison.
     * @return this Constraint to allow the chaining of method calls.
     */
    public Constraint not ();
    
    
    /**
     * returns the Object the query graph was constrained with to
     * create this Constraint.
     * @return Object the constraining object.
     */
    public Object getObject();

    /**
     * return only object that are distinct for this constraint.
     * If two object has the sa
     * @return this Constraint to allow the chaining of method calls.
     */
    public Constraint distinct ();

}
