/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
package org.openmobileis.database.fastobjectdb.db;

import java.io.IOException;
import java.util.Hashtable;

import org.openmobileis.database.fastobjectdb.*;
import org.openmobileis.database.fastobjectdb.db.crypto.FODBCypher;
import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndex;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndexFactory;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndexHeader;
import org.openmobileis.database.fastobjectdb.db.index.FODBUniqueIndex;
import org.openmobileis.database.fastobjectdb.db.index.node.Node;
import org.openmobileis.database.fastobjectdb.db.store.FODBCollectionFile;
import org.openmobileis.database.fastobjectdb.db.store.FODBCollectionIndexFile;

import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.exception.BadDataFormatException;

/**
 * 
 * FODBCollection is the class that represents each collection. It offers simple
 * functions that are mainly called by FastObjectDB. The unique index that index
 * collection object ids, must be added first.
 * 
 */
public class FODBCollection {
  private String name;

  private FastObjectDB database;

  private Class objectType;

  // private HashMap indexs;
  private Array indexs;

  private Hashtable indexsbyMember;

  private FODBCollectionFile collectionFile;

  private FODBCollectionIndexFile collectionFileLink;

  /**
   * The Constructor. It is private, it is called by local methods
   * createCollection and openCollection.
   * 
   * @param db -
   *          FastObjectDB instance that refers it.
   * @param colName -
   *          Collection Name.
   * @param cf -
   *          Collection File.
   * @param objType -
   *          Type of Objects stored in the collection.
   * @throws FODBException
   */
  private FODBCollection(FastObjectDB db, FODBCollectionDescriptor descriptor, FODBCollectionFile cf) throws BadDataFormatException, FODBException {
    name = descriptor.getCollectionName();
    database = db;
    collectionFile = cf;
    collectionFileLink = new FODBCollectionIndexFile(cf);
    indexs = new Array(10);
    indexsbyMember = new Hashtable();
    if (descriptor.getObjectType() != null) { // wich means we create the collection
      objectType = descriptor.getObjectType();
    } else { // We open an existing collection
      database.getTransactionManager().begin();
      database.getTransactionManager().enterTransaction(this.name);
      try {
        objectType = collectionFile.getObjectType();
        FODBIndexHeader[] indexsHeaders;
        indexsHeaders = collectionFile.getIndexsHeaders();

        if (indexsHeaders != null) {
          FODBIndex index;
          for (int i = 0; i < indexsHeaders.length; i++) {
            index = FODBIndexFactory.getFactory().createIndex(indexsHeaders[i], objectType, collectionFileLink);
            if (index == null) {
              throw new FODBException("Internal Error : Can't open collection !!!");
            }
            indexs.add(indexsHeaders[i].getMemberName());
            indexsbyMember.put(indexsHeaders[i].getMemberName(), index);
          }
        }
      } catch (Throwable ex) {
        throw new FODBException(ex);
      } finally {
        database.getTransactionManager().commit();
      }
    }
  }
  
  public FODBCollectionDescriptor getDescriptor() {
    return collectionFile.getDescriptor();
  }

  public Class getCollectionObjectClass() {
    return objectType;
  }

  public FODBIndex getIndexByMember(String member) {
    return (FODBIndex) indexsbyMember.get(member);
  }

  public boolean isSynchronized() {
    return collectionFile.isSynchronized();
  }

  /**
   * Function that is called to create a new collection.
   * 
   * @param db -
   *          FastObjectDB instance that refers to it.
   * @param colName -
   *          Collection name.
   * @param objType -
   *          Type of Objects stored in the collection.
   * @return the object created.
   * @throws FODBException
   */
  public static FODBCollection createCollection(FastObjectDB db, FODBCollectionDescriptor descriptor) throws FODBException {
    // TODO : v�rifier que l'object est bien Serializable ou Externalizable.
    String completeFilePath = "" + db.getRootDir() + "/" + db.getName() + "/" + descriptor.getCollectionName() + ".col";
    try {
      FODBCollectionFile fcf = FODBCollectionFile.createCollection(completeFilePath, descriptor, db);
      return new FODBCollection(db, descriptor, fcf);
    } catch (Throwable ex) {
      throw new FODBException(ex);
    }
  }

  /**
   * Function that is called to open an existing collection.
   * 
   * @param db -
   *          FastObjectDB instance that refers to it.
   * @param colName -
   *          Collection name.
   * @return the object created.
   * @throws FODBException
   */
  public static FODBCollection openCollection(FastObjectDB db, String colName) throws FODBException {
    String completeFilePath = "" + db.getRootDir() + "/" + db.getName() + "/" + colName + ".col";
    try {
      FODBCollectionFile fcf = FODBCollectionFile.openCollection(completeFilePath, colName, db);
      return new FODBCollection(db, new FODBCollectionDescriptor(colName, null), fcf);
    } catch (Throwable ex) {
      throw new FODBException(ex);
    }
  }

  public String getName() {
    return name;
  }

  /**
   * Init collection encryption with the specified cypher. Must be call just
   * after collection initialization before storing/reading any data. Same
   * cypher must be provided to read already stored object.
   * 
   * @param cypher :
   *          cypher to be use to encrypt data.
   * @throws FODBException
   */
  public void setCollectionCypher(FODBCypher cypher) throws FODBException {
    collectionFile.setCollectionCypher(cypher);
  }

  /**
   * return the collection cypher if any, null otherwise.
   * 
   * @return cypher use to encrypt data, null otherwise.
   */
  public FODBCypher getCollectionCypher() throws FODBException {
    return collectionFile.getCollectionCypher();
  }

  /**
   * Add an index.
   * 
   * @param descriptor -
   *          index descriptor.
   * @return true if it succeded, false otherwise.
   * @throws FODBException
   */
  public boolean addIndex(FODBIndexDescriptor descriptor) throws FODBException {
    String memberName = descriptor.getMemberName();
    if (indexs.contains(memberName)) {
      return false;
    }
    database.getTransactionManager().begin();
    try {
      FODBIndex newIndex = FODBIndexFactory.getFactory().createIndex(descriptor, objectType, collectionFileLink);

      if (newIndex == null) {
        System.out.println("" + descriptor.getName() + " ===> NULL !!!");
        return false;
      }

      indexs.add(memberName);
      indexsbyMember.put(memberName, newIndex);
      return true;
    } catch (Throwable ex) {
      throw new FODBException(ex);
    } finally {
      database.getTransactionManager().commit();
    }
  }

  public FODBIndex[] getCollectionIndexList() {
    FODBIndex[] ret = new FODBIndex[indexs.size()];
    for (int i = 0; i < indexs.size(); i++) {
      String memberName = (String) indexs.get(i);
      ret[i] = (FODBIndex) indexsbyMember.get(memberName);
    }
    return ret;
  }

  public FODBUniqueIndex getCollectionIdIndex() throws FODBException {
    if (indexs.size() == 0) {
      throw new FODBException("Internal collection error (index arrayList empty).");
    }
    String firstIndexMember = (String) indexs.get(0);
    FODBIndex firstIndex = (FODBIndex) indexsbyMember.get(firstIndexMember);
    if (firstIndex.getType() != FODBIndex.UNIQUE) {
      throw new FODBException("Trying to delete single element in a Collection which only contains multiple indexs. Use the query delete method to do so.");
    }

    // If we got there, our first index is a Unique-Index
    FODBUniqueIndex uIndex = (FODBUniqueIndex) firstIndex;
    return uIndex;
  }

  /**
   * Add an Element to the collection.
   * 
   * @param obj -
   *          object to add.
   * @return true if it succeded, false otherwise.
   * @throws FODBException
   */
  public boolean addElement(Object obj) throws FODBException {
    if (!objectType.isInstance(obj)) {
      throw new FODBException("Trying to add element of a wrong type.");
    }

    if (indexs.size() == 0) {
      throw new FODBException("Trying to add element to a collection which doesn't have any index.");
    }

    database.getTransactionManager().begin();
    try {
      long pos = collectionFile.writeObject(obj);

      for (int i = 0; i < indexs.size(); i++) {
        try {
          String memberName = (String) indexs.get(i);
          ((FODBIndex) indexsbyMember.get(memberName)).add(obj, pos);
        } catch (Exception ex) {
          throw new FODBException(ex);
        }
      }
    } catch (Throwable ex) {
      throw new FODBException(ex);
    } finally {
      database.getTransactionManager().commit();
    }

    return true;
  }

  /**
   * Replace an existing element of the Collection.
   * 
   * @param obj -
   *          object that will replace an existing one.
   * @return true if it succeded, false otherwise.
   * @throws FODBException
   */
  public boolean replaceElement(Object obj) throws FODBException {
    if (!objectType.isInstance(obj)) {
      throw new FODBException("Trying to add element of a wrong type.");
    }

    if (indexs.isEmpty()) {
      throw new FODBException("Trying to replace element in an empty collection (doesn't have any index).");
    }

    String memberName = (String) indexs.get(0);
    FODBIndex firstIndex = (FODBIndex) indexsbyMember.get(memberName);

    if (firstIndex.getType() != FODBIndex.UNIQUE) {
      throw new FODBException("Trying to replace an element in a collection which doesn't have any Unique Index !");
    }

    database.getTransactionManager().begin();
    try {
      // If we got there, our first index is a Unique-Index
      FODBUniqueIndex uIndex = (FODBUniqueIndex) firstIndex;
      Object key = ((FODBIndex) uIndex).getKey(obj);
      long elementPos = uIndex.getKeyPosition(key);

      if (elementPos == Node.NO_NODE) {
        return false;
      }
      this.deleteElementAtPos(elementPos);
      this.addElement(obj);
    } catch (Throwable ex) {
      throw new FODBException("Internal error !", ex);
    } finally {
      database.getTransactionManager().commit();
    }

    // everything went fine :)
    return true;
  }

  /**
   * Delete an element of the Collection.
   * 
   * @param obj -
   *          object to Delete.
   * @return - true if it succeded, false otherwise.
   * @throws FODBException
   */
  public boolean deleteSingleElement(Object obj) throws FODBException {
    if (!objectType.isInstance(obj)) {
      throw new FODBException("Trying to delete element of a wrong type.");
    }

    if (indexs.isEmpty()) {
      throw new FODBException("Trying to delete element from an empty collection (doesn't have any index).");
    }

    String memberName = (String) indexs.get(0);
    FODBIndex firstIndex = (FODBIndex) indexsbyMember.get(memberName);
    if ((firstIndex).getType() != FODBIndex.UNIQUE) {
      throw new FODBException("Trying to delete single element in a Collection which only contains multiple indexs. Use the query delete method to do so.");
    }

    database.getTransactionManager().begin();
    try {
      // If we got there, our first index is a Unique-Index
      FODBUniqueIndex uIndex = (FODBUniqueIndex) firstIndex;
      Object key = ((FODBIndex) uIndex).getKey(obj);
      this.deleteObjectWithId(key);
    } catch (Exception ex) {
      throw new FODBException("Internal error !! Database might be corrupted !!!", ex);
    } finally {
      database.getTransactionManager().commit();
    }

    return true;
  }

  /**
   * Delete an object of the Collection with its ID. The first index in the
   * collection is use as ID unique index.
   * 
   * @param id -
   *          id of object to Delete.
   * @return - object position or NODE.NO_NODE if not found.
   * @throws FODBException
   */
  public boolean deleteObjectWithId(Object id) throws FODBException {

    if (indexs.isEmpty()) {
      throw new FODBException("Trying to delete element from an empty collection (doesn't have any index).");
    }

    String memberName = (String) indexs.get(0);
    FODBIndex firstIndex = (FODBIndex) indexsbyMember.get(memberName);
    if ((firstIndex).getType() != FODBIndex.UNIQUE) {
      throw new FODBException("Trying to delete single element in a Collection which only contains multiple indexs. Use the query delete method to do so.");
    }

    database.getTransactionManager().begin();
    try {
      // If we got there, our first index is a Unique-Index
      FODBUniqueIndex uIndex = (FODBUniqueIndex) firstIndex;
      long elementPos = uIndex.getKeyPosition(id);

      if (elementPos == Node.NO_NODE) {
        return false;
      }

      Object obj = getElementAtPos(elementPos);

      for (int i = 0; i < indexs.size(); i++) {
        memberName = (String) indexs.get(i);
        FODBIndex index = (FODBIndex) indexsbyMember.get(memberName);
        index.delObjectWithKey(obj, elementPos);
        // index.delObjectWithKey(index.getKey(obj), elementPos);
        // ((FODBIndex) indexsbyMember.get(memberName)).delObjectWithKey(id,
        // elementPos);
      }
      collectionFile.deleteObject(elementPos);
    } catch (Exception ex) {
      throw new FODBException("Internal error !! Database might be corrupted !!!", ex);
    } finally {
      database.getTransactionManager().commit();
    }

    return true;
  }

  public ForwardUniqueIndexIterator getForwardIndexIterator(String memberName) throws FODBException {
    FODBIndex index = (FODBIndex) indexsbyMember.get(memberName);
    if (index == null)
      throw new FODBException("index member not found in this collection");
    try {
      return new ForwardUniqueIndexIterator(index, this);
    } catch (IOException ex) {
      throw new FODBException(ex);
    } catch (ClassNotFoundException ex) {
      throw new FODBException(ex);
    }
  }

  public BackwardUniqueIndexIterator getBackwardIndexIterator(String memberName) throws FODBException {
    FODBIndex index = (FODBIndex) indexsbyMember.get(memberName);
    if (index == null)
      throw new FODBException("index member not found in this collection");
    try {
      return new BackwardUniqueIndexIterator(index, this);
    } catch (IOException ex) {
      throw new FODBException(ex);
    } catch (ClassNotFoundException ex) {
      throw new FODBException(ex);
    }
  }

  /**
   * Delete an element of the collection, knowing its position in the file.
   * 
   * @param objPos -
   *          position of the object to delete.
   * @throws FODBException
   */
  private void deleteElementAtPos(long objPos) throws FODBException {
    long posTab[] = new long[1];
    posTab[0] = objPos;
    deleteElements(posTab);
  }

  /**
   * Delete a list of elements, knowing there positions in the file.
   * 
   * @param objPos -
   *          table of Object pointers in the file.
   * @throws FODBException
   */
  private void deleteElements(long[] objPos) throws FODBException {
    if (indexs.isEmpty()) {
      throw new FODBException("Trying to remove an element from an empty collection.");
    }

    Object objToDelete = null;
    for (int j = 0; j < objPos.length; j++) {
      objToDelete = getElementAtPos(objPos[j]);
      this.deleteSingleElement(objToDelete);
    }
  }

  /**
   * Returns the element at position 'pos'
   * 
   * @param pos -
   *          postion of the element to retrieve.
   * @return - the element retrieved.
   * @throws FODBException
   */
  public Object getElementAtPos(long pos) throws FODBException {
    Object result;
    try {
      database.getTransactionManager().validateThread(Thread.currentThread());
      result = collectionFile.readObject(pos);
      return result;
    } catch (Exception ex) {
      throw new FODBException(ex);
    }
  }

  /**
   * Returns the node at position 'pos'
   * 
   * @param pos -
   *          postion of the element to retrieve.
   * @return - the element retrieved.
   * @throws FODBException
   */
  protected Object getNodeAtPos(long pos) throws FODBException {
    Object result;
    try {
      result = collectionFile.readNode(pos);
      return result;
    } catch (Throwable ex) {
      throw new FODBException(ex);
    }
  }

  /**
   * @return
   */
  public FastObjectDB getDatabase() {
    return database;
  }

}
