/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */
package org.openmobileis.database.fastobjectdb.db;

import java.io.IOException;


import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndex;
import org.openmobileis.database.fastobjectdb.db.index.FODBUniqueIndex;
import org.openmobileis.database.fastobjectdb.db.index.SearchResult;
import org.openmobileis.database.fastobjectdb.db.index.node.Node;

/**
 * Use to enumerate an index without to loading in memory all index key or referenced object.
 * Only the current object is loaded.
 * Start at the beginning and enumerate forward.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class ForwardUniqueIndexIterator {
	private FODBCollection collection;
	private Node currentNode;
	private int position = 0;
	private boolean end = false;
	private FODBIndex index;
	private boolean inited  = false;

  /**
   * 
   */
  ForwardUniqueIndexIterator(FODBIndex index, FODBCollection collection) throws IOException, ClassNotFoundException, FODBException {
    super();
    if (!(index instanceof FODBUniqueIndex))	{
    	throw new FODBException("Not unique index to iterator awailable");
    }
    this.index = index;
    this.collection = collection;
  }  
  
  private void initIndex()  throws FODBException	{
		collection.getDatabase().getTransactionManager().begin();
		end = false;
		try	{
			collection.getDatabase().getTransactionManager().enterTransaction(this.collection.getName());
			this.currentNode = index.readRoot();
			if (this.currentNode.nbKey == 0)	{ //empty index
				end = true;
			}
    	
			//position to the first positon
			long childBranch = currentNode.branchs[this.position];
			while (childBranch != Node.NO_NODE)	{
				this.currentNode = (Node)collection.getNodeAtPos(childBranch);
				this.position = 0; 
				childBranch = currentNode.branchs[this.position];
			}
		} catch (Throwable ex)	{
			throw new FODBException(ex);
		} finally	{
			collection.getDatabase().getTransactionManager().commit();
		}
		inited  = true;
  }
  
  public void initCurrentNode(Object indexKey) throws FODBException	{
		inited  = false;
		collection.getDatabase().getTransactionManager().begin();
		try	{
			collection.getDatabase().getTransactionManager().enterTransaction(this.collection.getName());
  		SearchResult result = ((FODBUniqueIndex)index).getNodeForKey(indexKey);
  		if (result.found)	{
  			this.currentNode = result.node;
  			this.position = result.pos;
				end = false;
				inited  = true;
  		}
  		this.next(); //position to next record.
		} finally	{
			collection.getDatabase().getTransactionManager().commit();
		}
  	
  }
  
	public boolean hasNext() throws FODBException	{
		return !end;
	}

  public Object next() throws FODBException	{
  	if (!inited) this.initIndex();
  	if (end) return null;
		collection.getDatabase().getTransactionManager().begin();
  	try	{
			collection.getDatabase().getTransactionManager().enterTransaction(this.collection.getName());
			
			//read current obj
			long objPos= currentNode.ptr[this.position];
			Object ret = collection.getElementAtPos(objPos);
			
			//get next pos
			if (currentNode.branchs[(this.position+1)] != Node.NO_NODE)	{ //goes down
				long childBranch = currentNode.branchs[this.position+1];
				while (childBranch != Node.NO_NODE)	{
					this.currentNode = (Node)collection.getNodeAtPos(childBranch);
					this.position = 0; 
					childBranch = currentNode.branchs[this.position];
				}
			} else	{
				this.position ++;  //goes next
				while (this.position >= currentNode.nbKey)	{ //goes up
					long parent = currentNode.parentPtr;
					if (parent == Node.NO_NODE)	{ //root node end index traversal.
						end = true;
						break; 
					}
					long nodepos = currentNode.filePtr;
					this.currentNode = (Node)collection.getNodeAtPos(parent);
					//child node position
					int i=0;
					for (; i<this.currentNode.branchs.length; i++)	{
						if (nodepos == this.currentNode.branchs[i]) break;
					}
					this.position = i;
				}
			}
			
	  	return ret;
  	} finally	{
			collection.getDatabase().getTransactionManager().commit();
  	}
  	
  }

}
