/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
package org.openmobileis.database.fastobjectdb.db.store;

import java.io.*;

import org.openmobileis.database.fastobjectdb.db.index.node.Node;
import org.openmobileis.database.fastobjectdb.db.transaction.TransactionFile;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Low level classe that store Object in a file.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class FODBObjectDatabaseFile implements TransactionFile {
  //-------------------------
  // constants
  //-------------------------
  protected static final long DEL_LIST_END = -1l;
  protected static final byte IS_DELETED = 0;
  protected static final byte IS_ACTIVE = 1;

  //-------------------------
  // fields
  //-------------------------
  protected RandomAccessFile dataFile;
  protected String fileName;
  protected long firstDel = DEL_LIST_END;

  //-------------------------
  // constructors
  //-------------------------
  public FODBObjectDatabaseFile(String name) throws IOException {
    fileName = new String(name);
  }

  public FODBObjectDatabaseFile(File file) throws IOException {
    fileName = new String(file.getName());
  }

  protected void constructorHelper() throws IOException {
		dataFile.seek(0);
    if (dataFile.length() == 0) {
      dataFile.writeLong(DEL_LIST_END);
      firstDel = DEL_LIST_END;
    } else
      firstDel = dataFile.readLong();
  }

  //-------------------------
  // implement object I/O
  //-------------------------

  // write object to current position
  public long writeObject(Object obj) throws IOException {
    ByteArrayOutputStream bytes = new ByteArrayOutputStream();
    ObjectOutputStream ostrm = new ObjectOutputStream(bytes);
    ostrm.writeObject(obj);
    ostrm.flush();

    // get the length of the output data
    int datalen = bytes.size(); // data length
    int reclen = datalen; // rec length
    long pos; // beginning of rec header

    // find someplace to put the record
    if (firstDel == DEL_LIST_END) {
      pos = dataFile.length();
      dataFile.seek(pos);
    } else { // replace a deleted record
      // start with first deleted record
      long prev = DEL_LIST_END;

      pos = firstDel;

      while (true) {
        dataFile.seek(pos);

        // read header info
        byte deleted = dataFile.readByte();
        dataFile.readInt();
        reclen = dataFile.readInt();
        //LogManager.traceDebug(0,"objectDatabasefile write deleted write length :"+reclen);
        long next = dataFile.readLong();

        // check integrity
        if (deleted == IS_ACTIVE)
          throw new IOException("corrupt delete list");

        // is deleted record big enough for new data?
        if (reclen >= datalen) {
          if (prev == DEL_LIST_END) {
            firstDel = next;

            // write new deleted record # at beginning of file
            dataFile.seek(0);
            dataFile.writeLong(firstDel);
          } else { // go back to previous record
            // read and change prev header
            dataFile.seek(prev + 1 + 4 + 4);
            dataFile.writeLong(next);
          }

          dataFile.seek(pos);

          break;
        }

        // deleted record is too small; go to next (or end)
        prev = pos;

        // do we have any more deleted records in the chain?
        if (next == DEL_LIST_END) {
          pos = dataFile.length();
          dataFile.seek(pos);
          reclen = datalen; // set new rec length Added 19/11/2003 PhD
          //LogManager.traceDebug(0,"objectDatabasefile  WRITE NEW NEW");
          break;
        } else
          pos = next;
      }
    }

    // patch to update the store position of page
    if ((obj instanceof Node)) {
      ((Node) obj).filePtr = pos;
      bytes = new ByteArrayOutputStream();
      ostrm = new ObjectOutputStream(bytes);
      ostrm.writeObject(obj);
      ostrm.flush();
      // get the length of the output data
      
      // Only changed the filePtr ... Serialization size doesn't change.
      //datalen = bytes.size(); // data length
      //reclen = datalen; // rec length
    }
    
    dataFile.writeByte(IS_ACTIVE);
    dataFile.writeInt(datalen);
    dataFile.writeInt(reclen);

    /*LogManager.traceDebug(0,"objectDatabasefile  WRITE new write length rec :"+reclen+" data : "+datalen);
    if (reclen < datalen)	{
    LogManager.traceDebug(0,"objectDatabasefile  WRITE NEW NEW");
    }*/
    dataFile.writeLong(DEL_LIST_END);
    dataFile.write(bytes.toByteArray());

    return pos;
  }

  // write an object at the current position
  public void rewriteObject(Object obj) throws IOException {
    ByteArrayOutputStream bytes = new ByteArrayOutputStream();
    ObjectOutputStream ostrm = new ObjectOutputStream(bytes);
    ostrm.writeObject(obj);
    ostrm.flush();

    // get the length of the output data
    int datalen = bytes.size(); // data length
    
    //test to see if new record feet in the old record
    long pos = dataFile.getFilePointer();
    dataFile.skipBytes(5);
    int oldreclen = dataFile.readInt();
		dataFile.seek(pos); // Go back to the beginning of the record.
		
    if (oldreclen < datalen)	{
			throw new IOException("ERROR FODBObjectDatabaseFile : rewriteObject : object to big to be rewritten." + obj.getClass().getName());
    }

    // make sure this record is marked as not deleted
    dataFile.writeByte(IS_ACTIVE);

    // write the length of the output
    dataFile.writeInt(datalen);

    // write the length of the record
    //dataFile.writeInt(reclen);
    dataFile.skipBytes(4);

    // write the "deleted" pointer
    dataFile.writeLong(DEL_LIST_END);

    // store the object to the actual file
    dataFile.write(bytes.toByteArray());
  }

  // read object from current record
  public Object readObject() throws IOException, ClassNotFoundException {
    // check record status
    byte deleted = dataFile.readByte();

    // break from loop if this is an active record
    if (deleted != IS_ACTIVE)
      throw new IOException("Try to read where there is no record");
    // get the length of the object
    int datalen = dataFile.readInt();

    // get length of record
    int reclen = dataFile.readInt();

    // skip the "deleted" pointer
    dataFile.skipBytes(8);

    // retrieve the object data and store it
    byte[] data = new byte[datalen];
    dataFile.readFully(data);

    // position pointer to beginning of next record
    int diff = reclen - datalen;

    if (diff > 0)
      dataFile.skipBytes(diff);

    ByteArrayInputStream bytes = new ByteArrayInputStream(data);
    ObjectInputStream istrm = new ObjectInputStream(bytes);
    return istrm.readObject();
  }

  // delete the current object record
  public void delete() throws IOException {
    // assume beginning of record, get pointer
    long pos = dataFile.getFilePointer();

    // is this record already deleted?
    byte deleted = dataFile.readByte();

    if (deleted == IS_DELETED)
      return;

    // back up and write new marker
    dataFile.seek(pos);
    dataFile.writeByte(IS_DELETED);

    // skip data length value
    dataFile.skipBytes(4 + 4);
		dataFile.writeLong(firstDel);

    // now change file header
    firstDel = pos;
    dataFile.seek(0);
    dataFile.writeLong(firstDel);
  }  

  // go to beginning of file
  public void rewind() throws IOException {
    dataFile.seek(8);
  }

  // skip over an object record
  public void skip() throws IOException {
    // skip over parts of record header
    dataFile.skipBytes(1 + 4);

    // read record length
    int reclen = dataFile.readInt();

    // skip to beginning of next record
    dataFile.skipBytes(8 + reclen);
  }

  public String getFileName() {
    return fileName;
  }

  //--------------------------------------------
  // provide shells for RandomAccessFile methods
  //--------------------------------------------
  public final FileDescriptor getFD() throws IOException {
    return dataFile.getFD();
  }

  public long getFilePointer() throws IOException {
    return dataFile.getFilePointer();
  }

  public void seek(long pos) throws IOException {
    dataFile.seek(pos);
  }

  public long length() throws IOException {
    return dataFile.length();
  }
  
	public void close() throws IOException {
			if (dataFile != null)    {
				dataFile.close();
				dataFile = null;
			}
	}
  
	public void open() throws IOException	{
		try {			
			dataFile = new RandomAccessFile(fileName, "rw");
			constructorHelper();
		} catch (Throwable ex) { // if an error occurs close the datafile.
			try {
				if (dataFile != null) {
					dataFile.close();
				}
			} catch (Throwable exi) {
				org.openmobileis.common.util.log.LogManager.traceError(0, "ObjectDatabaseFile constructorHelper error :" + exi.getMessage());
			}
			throw new IOException(ex.getMessage());
		}
	}
}
