/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.database.fastobjectdb.synchro.client;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectOutputStream;

import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.common.util.file.FileUtilities;
import org.openmobileis.common.util.file.ZipUtilities;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.database.DatabaseException;
import org.openmobileis.database.fastobjectdb.FODBCollectionDescriptor;
import org.openmobileis.database.fastobjectdb.FODBIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.FastObjectDBManager;
import org.openmobileis.database.fastobjectdb.db.FODBCollection;
import org.openmobileis.database.fastobjectdb.db.crypto.FODBCypher;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndex;
import org.openmobileis.synchro.journal.JournalLogRenderer;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.client.OpenMSPDBSyncListener;
import org.openmobileis.synchro.openmsp.client.db.ImportQueryManager;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: 
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class OpenMSPFODBSyncListener extends OpenMSPDBSyncListener {

  private String collectionName;

  private FODBImportQueryManager queryManager;

  /**
   * 
   */
  public OpenMSPFODBSyncListener(String collectionName) {
    super();
    this.collectionName = collectionName;
    queryManager = new FODBImportQueryManager(this.collectionName);
  }

  public void registerSynchroListener(SynchroFODBReturnListener listener) {
    this.queryManager.registerSynchroListener(listener);
  }

  /**
   * Construct collection description to be use by server to generate collection files.
   */
  protected String getSendCommandMetaData() {
    try {
      FODBCollection collection = ((SynchroFastObjectDB) FastObjectDBManager.getCurrentFODB()).getCollection(this.getSyncName());
      StringBuffer buff = new StringBuffer(collection.getCollectionObjectClass().getName());
      //add cypher if any
      FODBCypher cypher = collection.getCollectionCypher();
      if (cypher == null) {
        buff.append(":nocypher");
      } else {
        ByteArrayOutputStream bstr = new ByteArrayOutputStream();
        ObjectOutputStream ostr = new ObjectOutputStream(bstr);
        ostr.writeObject(cypher);
        buff.append(":");
        buff.append(new String(GeneralCoder.encodeBase64(bstr.toByteArray())));
      }
      
      //manage collection descriptor
      FODBCollectionDescriptor coldesc = collection.getDescriptor();
      try {
        ByteArrayOutputStream bstr = new ByteArrayOutputStream();
        ObjectOutputStream ostr = new ObjectOutputStream(bstr);
        ostr.writeObject(coldesc);
        byte[] b = bstr.toByteArray();
        buff.append(":");
        buff.append(new String(GeneralCoder.encodeBase64(b)));
      } catch (Throwable ex) {
        LogManager.traceError(0, ex);
      }

      FODBIndex[] indexList = collection.getCollectionIndexList();
      for (int i = 0; i < indexList.length; i++) {
        FODBIndexDescriptor descriptor = indexList[i].getIndexDescriptor();
        try {
          ByteArrayOutputStream bstr = new ByteArrayOutputStream();
          ObjectOutputStream ostr = new ObjectOutputStream(bstr);
          ostr.writeObject(descriptor);
          byte[] b = bstr.toByteArray();
          buff.append(":");
          buff.append(new String(GeneralCoder.encodeBase64(b)));
        } catch (Throwable ex) {
          LogManager.traceError(0, ex);
        }
      }
      return buff.toString();
    } catch (Throwable ex) {
      LogManager.traceError(0, ex);
    }
    return null;
  }

  /**
   * Call during complete synchronisation to dump the database with server data store in
   * the specified file name.
   * @param importFile : file containing server dump data file.
   */
  protected void importCompletSynchroDatabaseFile(String importFile) throws DatabaseException {
    this.replaceDBFiles(importFile);

    //reload db to take into account file change
    SynchroFastObjectDBManager.flushDB();
  }

  private void replaceDBFiles(String importFile) throws DatabaseException {
    FastObjectDB db = FastObjectDBManager.getCurrentFODB();
    String dbPath = "" + db.getRootDir() + "/" + db.getName()+ "/";
    importFile = FileUtilities.convertFileNameToSystem(importFile);
    try {
      BufferedInputStream inst = new BufferedInputStream(new FileInputStream(importFile));
      try {
        ZipUtilities.unZipData(inst, dbPath);
      } finally {
        inst.close();
        File filein = new File(importFile);
        filein.delete();
      }
    } catch (Throwable ex) {
      throw new DatabaseException(ex);
    }
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPDBSyncListener#getDBDataForAdd(java.lang.String)
   */
  protected String[] getDBDataForAdd(String pdaUID) throws OpenMSPException {
    String[] ret = null;
    try {
      Object obj = ((SynchroFastObjectDB) FastObjectDBManager.getCurrentFODB()).getObjectFromCollectionWithId(this.getSyncName(), pdaUID);
      ret = new String[2];
      ret[0] = pdaUID;
      ByteArrayOutputStream bstr = new ByteArrayOutputStream();
      ObjectOutputStream ostr = new ObjectOutputStream(bstr);
      ostr.writeObject(obj);
      byte[] b = bstr.toByteArray();
      ret[1] = new String(GeneralCoder.encodeBase64(b));
    } catch (Throwable ex) {
      throw new OpenMSPException(ex);
    }
    return ret;
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPDBSyncListener#getDBDataForReplace(java.lang.String)
   */
  protected String[] getDBDataForReplace(String pdaUID) throws OpenMSPException {
    return this.getDBDataForAdd(pdaUID);
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPDBSyncListener#getSynchroTarget()
   */
  protected String getSynchroTarget() {
    return collectionName;
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#getSyncName()
   */
  public String getSyncName() {
    return collectionName;
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPDBSyncListener#getSyncroCredential()
   */
  protected Credential getSyncroCredential() {
    return null;
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPDBSyncListener#getImportQueryManager()
   */
  protected ImportQueryManager getImportQueryManager() {
    return queryManager;
  }

  protected JournalLogRenderer getJournalLogRenderer() {
    return null;
  }

}
