/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */

package org.openmobileis.embedded.webserver;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;

import javax.servlet.ServletInputStream;

public class ServeInputStream extends ServletInputStream {

    private BufferedInputStream in;
    private int chunksize = 0;
    private boolean chunking = false;
    private boolean returnedAsReader, returnedAsStream;

    /* ------------------------------------------------------------ */
    /** Constructor
    */
    public ServeInputStream(InputStream in) {
      this.in = new BufferedInputStream(in);
    }

    /* ------------------------------------------------------------ */
    /** 
    * @param chunking 
    */
    public void chunking(boolean chunking) {
      this.chunking = chunking;
    }

    /* ------------------------------------------------------------ */
    /** Read a line ended by CR or CRLF or LF.
    * More forgiving of line termination than ServletInputStream.readLine().
    * This method only read raw data, that may be chunked.  Calling
    * ServletInputStream.readLine() will always return unchunked data.
    */
    // TODO: won't work with encoding
    public String readLine() throws IOException {
      StringBuffer buf = new StringBuffer(1024);

      int c;
      boolean cr = false;

      LineLoop : while ((c = chunking ? read() : in.read()) != -1) {
        switch (c) {
          case 10 :
            break LineLoop;

          case 13 :
            cr = true;
            if (!chunking)
              in.mark(2);
            break;

          default :
            if (cr) {
              //if (chunking)
              //log("Cannot handle CR in chunking mode");
              in.reset();
              break LineLoop;
            } else
              buf.append((char) c);
            break;
        }
      }

      if (c == -1 && buf.length() == 0)
        return null;
      //System.err.println(buf.toString()); //###

      return buf.toString();
    }

    /* ------------------------------------------------------------ */
    public int read() throws IOException {
      if (chunking) {
        int b = -1;
        if (chunksize <= 0 && getChunkSize() <= 0)
          return -1;
        b = in.read();
        chunksize = (b < 0) ? -1 : (chunksize - 1);
        return b;
      }

      return in.read();
    }

    /* ------------------------------------------------------------ */
    public int read(byte b[]) throws IOException {
      return read(b, 0, b.length);
    }

    /* ------------------------------------------------------------ */
    public int read(byte b[], int off, int len) throws IOException {
      if (chunking) {
        if (chunksize <= 0 && getChunkSize() <= 0)
          return -1;
        if (len > chunksize)
          len = chunksize;
        len = in.read(b, off, len);
        chunksize = (len < 0) ? -1 : (chunksize - len);
      } else
        len = in.read(b, off, len);
      //System.err.println(new String(b,off,len)); //###        

      return len;
    }

    /* ------------------------------------------------------------ */
    public long skip(long len) throws IOException {
      if (chunking) {
        if (chunksize <= 0 && getChunkSize() <= 0)
          return -1;
        if (len > chunksize)
          len = chunksize;
        len = in.skip(len);
        chunksize = (len < 0) ? -1 : (chunksize - (int) len);
      } else
        len = in.skip(len);
      return len;
    }

    /* ------------------------------------------------------------ */
    /** Available bytes to read without blocking.
    * If you are unlucky may return 0 when there are more
    */
    public int available() throws IOException {
      if (chunking) {
        int len = in.available();
        if (len <= chunksize)
          return len;
        return chunksize;
      }

      return in.available();
    }

    /* ------------------------------------------------------------ */
    public void close() throws IOException {
      in.close();
      chunksize = -1;
    }

    /* ------------------------------------------------------------ */
    /** Mark is not supported
    * @return false
    */
    public boolean markSupported() {
      return false;
    }

    /* ------------------------------------------------------------ */
    /** Not Implemented
    */
    public void reset() {
    }

    /* ------------------------------------------------------------ */
    /** Not Implemented
    * @param readlimit 
    */
    public void mark(int readlimit) {
    }

    /* ------------------------------------------------------------ */
    private int getChunkSize() throws IOException {
      if (chunksize < 0)
        return -1;

      chunksize = -1;

      // Get next non blank line
      chunking = false;
      String line = readLine();
      while (line != null && line.length() == 0)
        line = readLine();
      chunking = true;

      // Handle early EOF or error in format
      if (line == null)
        return -1;

      // Get chunksize
      int i = line.indexOf(';');
      if (i > 0)
        line = line.substring(0, i).trim();
      chunksize = Integer.parseInt(line, 16);

      // check for EOF
      if (chunksize == 0) {
        chunksize = -1;
        // Look for footers
        chunking = false;
      }
      return chunksize;
    }

    boolean isReturnedAsStream() {
      return returnedAsStream;
    }

    void setReturnedAsStream(boolean _on) {
      returnedAsStream = _on;
    }

    boolean isReturnedAsReader() {
      return returnedAsReader;
    }

    void setReturnedAsReader(boolean _on) {
      returnedAsReader = _on;
    }


}
