/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */

package org.openmobileis.embedded.webserver;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.servlet.RequestDispatcher;
import javax.servlet.Servlet;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.context.Plateform;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.file.FileUtilities;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;
import org.openmobileis.embedded.util.SystemAPI;

public class WebServer implements ServletContext, RequestDispatcher {
    
    private final static int DEFAULT_PORT = 8080;
    
    private ServerThreadPool connexionPool;
    
    protected PathTreeDictionary registry;
    // loaded servlets : ClassName -> servlet instance
    private Hashtable servlets = new Hashtable();
    
    private Hashtable attributes;
    
    boolean running = true;
    private ServerSocket serverSocket;
    protected String hostName;
    protected int port = DEFAULT_PORT;
    
    private static Properties webserverProps;
    
    // path for properties
    public static String servletFile = "";
    public static String installPath = "";

    
    /// Constructors.
    
    public WebServer() {
        this(System.getProperty("user.dir"));
    }
    
    public WebServer(String pathToConfFile) {
    	System.setProperty("user.dir", pathToConfFile);
        try {
            if (pathToConfFile != null) {
                webserverProps = PropertiesManager.getManager().getPropertiesFromPath(pathToConfFile+File.separator+"WEB-INF"+File.separator+"conf"+java.io.File.separator+"properties"+java.io.File.separator+"webserver.properties");
            } else {
                webserverProps = PropertiesManager.getManager().getPropertiesFromPath("properties"+java.io.File.separator+"webserver.properties");
            }
            
        } catch (Exception ex) {
            System.out.println("Unable to find webserver.properties !!\nexit(1)");
            System.exit(1);
        }
        String strPort = webserverProps.getProperty("port");
        if (strPort != null && strPort.length() > 0) {
            try {
                this.port = Integer.parseInt(strPort);
            } catch (NumberFormatException ex) {
                this.port = DEFAULT_PORT;
            }
        }
        
        WebServer.installPath = webserverProps.getProperty("installPath");
        servletFile = webserverProps.getProperty("servletFile");
        registry = new PathTreeDictionary();
        try {
            connexionPool = new ServerThreadPool();
        } catch (Exception ex) {
            LogManager.traceError(LogServices.WEBSERVICE,
                    "can not initialise server connexion pool");
            LogManager.traceError(0, ex);
        }
        attributes = new Hashtable();
    }
    
    /// Register a Servlet by class name.  Registration consists of a URL
    // pattern, which can contain wildcards, and the class name of the Servlet
    // to launch when a matching URL comes in.  Patterns are checked for
    // matches in the order they were added, and only the first match is run.
    public void addServlet(String urlPat, String className) {
      addServlet(urlPat, className, (Hashtable) null);
    }

    public void addServlet(String urlPat, String className, Hashtable initParams) {
      // Check if we're allowed to make one of these.
      SecurityManager security = System.getSecurityManager();
      if (security != null) {
        int i = className.lastIndexOf('.');
        if (i != -1) {
          security.checkPackageAccess(className.substring(0, i));
          security.checkPackageDefinition(className.substring(0, i));
        }
      }

      // Make a new one.
      try {
        addServlet(urlPat, (Servlet) Class.forName(className).newInstance(), initParams);
        return;
      } catch (ClassNotFoundException e) {
        LogManager.traceError(0, e);
        LogManager.traceError(LogServices.WEBSERVICE, "Class not found: " + className);
      } catch (ClassCastException e) {
        LogManager.traceError(0, e);
        LogManager.traceError(LogServices.WEBSERVICE, "Class cast problem: " + e.getMessage());
      } catch (InstantiationException e) {
        LogManager.traceError(0, e);
        LogManager.traceError(LogServices.WEBSERVICE, "Instantiation problem: " + e.getMessage());
      } catch (IllegalAccessException e) {
        LogManager.traceError(0, e);
        LogManager.traceError(LogServices.WEBSERVICE, "Illegal class access: " + e.getMessage());
      } catch (Exception e) {
        LogManager.traceError(0, e);
        LogManager.traceError(LogServices.WEBSERVICE, "Unexpected problem creating servlet: " + e);
        LogManager.traceError(LogServices.WEBSERVICE, e);
      }
    }

    /// Register a Servlet.  Registration consists of a URL pattern,
    // which can contain wildcards, and the Servlet to
    // launch when a matching URL comes in.  Patterns are checked for
    // matches in the order they were added, and only the first match is run.
    public void addServlet(String urlPat, Servlet servlet) {
      addServlet(urlPat, servlet, (Hashtable) null);
    }

    public void addServlet(String urlPat, Servlet servlet, Hashtable initParams) {
      try {
        servlet.init(new ServeConfig((ServletContext) this, initParams, urlPat));
        registry.put(urlPat, servlet);
        servlets.put(servlet.getClass().getName(), urlPat);
      } catch (ServletException e) {
          LogManager.traceError(0, e);
          LogManager.traceError(LogServices.WEBSERVICE,
                  "Problem initializing servlet: " + e);
      }
    }
    
    public Object[] getServletByURLorClassName(String name) {
        Object[] servlet = new Object[2];
        servlet = registry.get(name);
        if (servlet[0] == null) {
            String servletUri = (String)servlets.get(name);
            if (servletUri != null && servletUri.trim().length() > 0) {
                servlet = registry.get(servletUri);
            }
        }
        if (servlet[0] == null) {
            String newName = name;
            int index = newName.lastIndexOf('/');
            while (index != -1) {
                newName = newName.substring(0, index);
                servlet = registry.get(newName);
                if (servlet[0] != null) {
                    break;
                }
                index = newName.lastIndexOf('/');
            }
        }
        return servlet;
    }
    
    // Run the server.  Returns only on errors.
    public void serve() throws IOException{
        // Load Servlets
        //this.loadServlets();
    	
    	// First of all, instantiate the LogManager to the console
    	if (LogManager.getInstance() == null) {
    		LogManager.registerLogManager(null);
    	}
    	
        this.readServlets(servletFile);
        serverSocket = new ServerSocket(port, 1000);
        hostName = serverSocket.getInetAddress().getHostName();
        
        
        // start Web Browser
		try {
	    Plateform plateform = ApplicationContextManager.getManager().getApplicationContext().getPlateform();
	    if (plateform.getOS().equals(Plateform.POCKETPCOS))	{
	      if ((PropertiesManager.getManager().getProperty("org.openmobileis.embedded.execprog.browser")!=null) &&
	          (PropertiesManager.getManager().getProperty("org.openmobileis.embedded.execprog.browser.args")!=null))	{
	        System.loadLibrary("ExecDLL2002");
		      LogManager.traceInfo(0,"Start Pocket IE");
	        SystemAPI.execProgram(PropertiesManager.getManager().getProperty("org.openmobileis.embedded.execprog.browser")
						, PropertiesManager.getManager().getProperty("org.openmobileis.embedded.execprog.browser.args")+"/"+Long.toString(System.currentTimeMillis())); //add long to avoid IE cache
	      } else	{
		      LogManager.traceInfo(0,"Pocket IE not started on PocketPC no properties defined.");
	      }
			}
		} catch (Throwable e) {
			LogManager.traceError(0, e);
			LogManager.traceError(LogServices.WEBSERVICE, "Can not launch ie with parameters : " + PropertiesManager.getManager().getProperty("execprog.browser")
					+ " args : " + PropertiesManager.getManager().getProperty("execprog.browser.args"));
		}
        
        try { // accept request from clients
            while (true) {
                Socket socket = serverSocket.accept();
                connexionPool.runServerConnexion(this, socket);
            } // end while
        } catch (IOException e) {
            LogManager.traceError(0, e);
            LogManager.traceError(LogServices.WEBSERVICE,
                    "IOEXception during servlet process Accept: " + e);
        } finally {
            try {
                serverSocket.close();
            } catch (Exception e) {
                LogManager.traceError(LogServices.WEBSERVICE, e);
            }
        }
    }
    
    /// Destroys all currently-loaded servlets.
    public void destroyAllServlets() {
      Enumeration en = registry.elements();
      while (en.hasMoreElements()) {
        Servlet servlet = (Servlet) en.nextElement();
        servlet.destroy();
      }
      registry = new PathTreeDictionary();
      // invalidate all sessions?
    }
    
    /*
     * free all resources
     */
    public void freeResources() throws Exception {
        this.destroyAllServlets();
        connexionPool.freeResources();
        serverSocket.close();
        LogManager.traceInfo(LogServices.WEBSERVICE, "Web server is stopped");
        //System.exit(0);
    }
    
    private void readServlets(String cfgfile) {
        /** servlet.properties file format
        servlet.<servletname>.code=<servletclass>
        servlet.<servletname>.initArgs=<name=value>,<name=value>
            */
        Hashtable servletstbl, parameterstbl;
        servletstbl = new Hashtable();
        parameterstbl = new Hashtable();
        if (cfgfile != null) {
          File file = new File(cfgfile);
          if (file.exists() && file.canRead()) {
            try {
              DataInputStream in = new DataInputStream(new FileInputStream(file));
              /** format of servlet.cfg file
              servlet_name;servlet_class;init_parameter1=value1;init_parameter2=value2...
              */
              do {
                String servletdsc = in.readLine();
                if (servletdsc == null)
                  break;
                StringTokenizer dsctokenzr = new StringTokenizer(servletdsc, ".=,", false);
                if (dsctokenzr.hasMoreTokens()) {
                  if (!dsctokenzr.nextToken().equalsIgnoreCase("servlet")) {
                    System.err.println("No leading 'servlet' keyword, the sentence is skipped");
                    break;
                  }
                  if (dsctokenzr.hasMoreTokens()) {
                    String servletname = dsctokenzr.nextToken();

                    if (dsctokenzr.hasMoreTokens()) {
                      String lt = dsctokenzr.nextToken();
                      if (lt.equalsIgnoreCase("code")) {
                        if (dsctokenzr.hasMoreTokens())
                          servletstbl.put(servletname, dsctokenzr.nextToken("="));
                      } else if (lt.equalsIgnoreCase("initArgs")) {
                        Hashtable initparams = new Hashtable();
                        while (dsctokenzr.hasMoreTokens()) {
                          String key = dsctokenzr.nextToken("=,");
                          if (dsctokenzr.hasMoreTokens())
                            initparams.put(key, dsctokenzr.nextToken(",="));
                        }
                        parameterstbl.put(servletname, initparams);
                      } else
                        System.err.println("Unrecognized token " + lt + " in " + servletdsc + ", the line's skipped");
                    }
                  }
                }
              }
              while (true);
            } catch (IOException e) {
              System.err.println("Problem reading cfg file: " + e);
            }
            Enumeration se = servletstbl.keys();
            String servletname;
            while (se.hasMoreElements()) {
              servletname = (String) se.nextElement();
              addServlet(servletname, (String) servletstbl.get(servletname), (Hashtable) parameterstbl.get(servletname));
            }
          }
        }
      }
    
/*    protected void loadServlets() {
        // load servlets
        
        // add the FileServlet by default. By default, if there is no matching
        // with the other servlet,
        // the file servlet is choosen
        this.addServlet("/FileServlet", "org.openmobileis.services.servlet.FileServlet");
        
        java.util.Properties props = new java.util.Properties();
        try {
            props.load(new java.io.FileInputStream(servletFile));
            java.util.Enumeration enum = props.propertyNames();
            while (enum.hasMoreElements()) {
                String className = (String) enum.nextElement();
                String pattern = (String) props.getProperty(className);
                this.addServlet(pattern, className);
            }
        } catch (IOException ex) {
            LogManager.traceError(0, ex);
        }
    }*/
    
    
    
    
    
    
    // ServletContext METHODS
    // ***************************************************************************************************************

    /// Returns the value of the named attribute of the network service, or
    // null if the attribute does not exist.  This method allows access to
    // additional information about the service, not already provided by
    // the other methods in this interface.
    public Object getAttribute(String name) {
      // This server does not support attributes.
      return attributes.get(name);
    }
    
    public Enumeration getAttributeNames() {
        return attributes.keys();
	}
    
    public ServletContext getContext(String uripath) {
        return this; // only root context supported
	}
    
    // no way to specify parameters for context
    public String getInitParameter(String param) {
      return null;
    }
    
    public Enumeration getInitParameterNames() {
        return null;
	}
    
    public int getMajorVersion() {
        return 2; // support 2.x
	}

    /// Returns the MIME type of the specified file.
    // @param file file name whose MIME type is required
    public String getMimeType(String file) {
      file = file.toUpperCase();

      if (file.endsWith(".HTML") || file.endsWith(".HTM"))
        return "text/html";
      if (file.endsWith(".TXT"))
        return "text/plain";
      if (file.endsWith(".XML"))
        return "text/xml";
      if (file.endsWith(".CSS"))
        return "text/css";
      if (file.endsWith(".SGML") || file.endsWith(".SGM"))
        return "text/x-sgml";
      // Image
      if (file.endsWith(".GIF"))
        return "image/gif";
      if (file.endsWith(".JPG") || file.endsWith(".JPEG") || file.endsWith(".JPE"))
        return "image/jpeg";
      if (file.endsWith(".PNG"))
        return "image/png";
      if (file.endsWith(".TIF") || file.endsWith(".TIFF"))
        return "image/tiff";
      if (file.endsWith(".RGB"))
        return "image/x-rgb";
      if (file.endsWith(".XPM"))
        return "image/x-xpixmap";
      if (file.endsWith(".XBM"))
        return "image/x-xbitmap";
      if (file.endsWith(".SVG"))
        return "image/svg-xml ";
      if (file.endsWith(".SVGZ"))
        return "image/svg-xml ";
      // Audio
      if (file.endsWith(".AU") || file.endsWith(".SND"))
        return "audio/basic";
      if (file.endsWith(".MID") || file.endsWith(".MIDI") || file.endsWith(".RMI") || file.endsWith(".KAR"))
        return "audio/mid";
      if (file.endsWith(".MPGA") || file.endsWith(".MP2") || file.endsWith(".MP3"))
        return "audio/mpeg";
      if (file.endsWith(".WAV"))
        return "audio/wav";
      if (file.endsWith(".AIFF") || file.endsWith(".AIFC"))
        return "audio/aiff";
      if (file.endsWith(".AIF"))
        return "audio/x-aiff";
      if (file.endsWith(".RA"))
        return "audio/x-realaudio";
      if (file.endsWith(".RPM"))
        return "audio/x-pn-realaudio-plugin";
      if (file.endsWith(".RAM"))
        return "audio/x-pn-realaudio";
      if (file.endsWith(".SD2"))
        return "audio/x-sd2";
      // Application
      if (file.endsWith(".BIN") || file.endsWith(".DMS") || file.endsWith(".LHA") || file.endsWith(".LZH") || file.endsWith(".EXE") || file.endsWith(".CLASS"))
        return "application/octet-stream";
      if (file.endsWith(".HQX"))
        return "application/mac-binhex40";
      if (file.endsWith(".PS") || file.endsWith(".AI") || file.endsWith(".EPS"))
        return "application/postscript";
      if (file.endsWith(".PDF"))
        return "application/pdf";
      if (file.endsWith(".RTF"))
        return "application/rtf";
      if (file.endsWith(".DOC"))
        return "application/msword";
      if (file.endsWith(".PPT"))
        return "application/powerpoint";
      if (file.endsWith(".FIF"))
        return "application/fractals";
      if (file.endsWith(".P7C"))
        return "application/pkcs7-mime";
      // Application/x
      if (file.endsWith(".JS"))
        return "application/x-javascript";
      if (file.endsWith(".Z"))
        return "application/x-compress";
      if (file.endsWith(".GZ"))
        return "application/x-gzip";
      if (file.endsWith(".TAR"))
        return "application/x-tar";
      if (file.endsWith(".TGZ"))
        return "application/x-compressed";
      if (file.endsWith(".ZIP"))
        return "application/x-zip-compressed";
      if (file.endsWith(".DIR") || file.endsWith(".DCR") || file.endsWith(".DXR"))
        return "application/x-director";
      if (file.endsWith(".DVI"))
        return "application/x-dvi";
      if (file.endsWith(".TEX"))
        return "application/x-tex";
      if (file.endsWith(".LATEX"))
        return "application/x-latex";
      if (file.endsWith(".TCL"))
        return "application/x-tcl";
      if (file.endsWith(".CER") || file.endsWith(".CRT") || file.endsWith(".DER"))
        return "application/x-x509-ca-cert";
      // Video
      if (file.endsWith(".MPG") || file.endsWith(".MPE") || file.endsWith(".MPEG"))
        return "video/mpeg";
      if (file.endsWith(".QT") || file.endsWith(".MOV"))
        return "video/quicktime";
      if (file.endsWith(".AVI"))
        return "video/x-msvideo";
      if (file.endsWith(".MOVIE"))
        return "video/x-sgi-movie";
      // Chemical
      if (file.endsWith(".PDB") || file.endsWith(".XYZ"))
        return "chemical/x-pdb";
      // X-
      if (file.endsWith(".ICE"))
        return "x-conference/x-cooltalk";
      if (file.endsWith(".WRL") || file.endsWith(".VRML"))
        return "x-world/x-vrml";
      if (file.endsWith(".WML"))
        return "text/vnd.wap.wml";
      if (file.endsWith(".WMLC"))
        return "application/vnd.wap.wmlc";
      if (file.endsWith(".WMLS"))
        return "text/vnd.wap.wmlscript";
      if (file.endsWith(".WMLSC"))
        return "application/vnd.wap.wmlscriptc";
      if (file.endsWith(".WBMP"))
        return "image/vnd.wap.wbmp";

      return null;
    }
    
	public int getMinorVersion() {
	    return 3; // support 2.3
	}
    
    public RequestDispatcher getNamedDispatcher(String name) {
        return this;
    }
    
    public String getRealPath(String path) {
        return installPath + FileUtilities.convertFileNameToSystem(path);
        //return htdocs + File.separator + FileUtilities.convertFileNameToSystem(path);
    }
    
    public RequestDispatcher getRequestDispatcher(String path) {
        return this;
    }
    
    // only root relative in this implementation
    public URL getResource(String path) throws MalformedURLException {
      return new URL("http", hostName, port, path);
    }
    
    public InputStream getResourceAsStream(String path) {
        return null; // we don't provide resources in this way
	}
    
    public java.util.Set getResourcePaths(java.lang.String path) {
        // TODO: implement
        return null;
	}
    
    /// Returns the name and version of the web server under which the servlet
    // is running.
    // Same as the CGI variable SERVER_SOFTWARE.
    public String getServerInfo() {
      return Identification.serverName + " " + Identification.serverVersion + " (" + Identification.serverUrl + ")";
    }
    
    public Servlet getServlet(String name) {
        // Deprecated. As of Java Servlet API 2.1, with no direct replacement.
        // This method was originally defined to retrieve a servlet from a ServletContext.
        // In this version, this method always returns null and remains only to preserve binary compatibility.
        // This method will be permanently removed in a future version of the Java Servlet API.

        // In lieu of this method, servlets can share information using the ServletContext class
        // and can perform shared business logic by invoking methods on common non-servlet classes.
        return null;
    }
    
    /**
     * Returns the name of this web application correponding to this ServletContext as specified in the deployment descriptor for this web application by the display-name element.
       * @return The name of the web application or null if no name has been declared in the deployment descriptor.
     *
     * @since Servlet 2.3 
     */
    public java.lang.String getServletContextName() {
      return null; //"ROOT";
    }
    
    /// Enumerates the names of the servlets in this context (server). Only
    // servlets that are accesible will be returned.  This enumeration always
    // includes the servlet itself.
    public Enumeration getServletNames() {
        // Deprecated. As of Java Servlet API 2.1, with no replacement.
        // This method was originally defined to return an Enumeration of all the servlet names known to this context.
        // In this version, this method always returns an empty Enumeration and remains only to preserve binary compatibility.
        // This method will be permanently removed in a future version of the Java Servlet API.
        return new Vector().elements();
	}
    
    public Enumeration getServlets() {
        // Deprecated. As of Java Servlet API 2.0, with no replacement.
        // This method was originally defined to return an Enumeration of all the servlets known to this servlet context.
        // In this version, this method always returns an empty enumeration and remains only to preserve binary compatibility.
        // This method will be permanently removed in a future version of the Java Servlet API.
        return new Vector().elements();
    }
    
    public void log(Exception exception, String msg) {
        LogManager.traceWarning(LogServices.WEBSERVICE, msg);
        LogManager.traceWarning(LogServices.WEBSERVICE, exception);
    }
    
    public void log(String msg) {
        LogManager.traceWarning(LogServices.WEBSERVICE, msg);
    }
    
    public void log(String message, Throwable throwable) {
        LogManager.traceWarning(LogServices.WEBSERVICE, message);
        LogManager.traceWarning(LogServices.WEBSERVICE, throwable);
    }
    
    public void removeAttribute(String name) {
        attributes.remove(name);
	}
    
    public void setAttribute(String name, Object object) {
        attributes.put(name, object);
	}
    
    
    
    // RequestDispatcher METHODS
    // ***************************************************************************************************************
    
    public void forward(ServletRequest request, ServletResponse response) {
    }
    public void include(ServletRequest request, ServletResponse response) {
    }
}
