/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */

package org.openmobileis.embedded.webserver.templates;

import freemarker.template.*;

import java.io.*;

import javax.servlet.http.HttpServletResponse;

import org.openmobileis.common.intl.IntlResourceManager;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.services.common.ServletTools;
import org.openmobileis.common.util.log.*;
import org.openmobileis.common.util.exception.ServerException;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Manage pattern (templates) for HTML forms used by the PDA web server<p>
 * It is based on singleton model.
 * The path gives the location of the patterns (should terminate by a file separator).
 * The API getHTMLFile (patternName, variables) allow to get the HTML file from a pattern
 * and a list of variables.
 * The default implementation use the freemarker template manager. See freemarker
 * for more information.
 * Inside the pattern, the variables are tagged by ${}.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */

public class TemplateManager  implements CacheListener {

 // private static FileTemplateCache templateCache;

  private static TemplateManager manager = null;
  private String rootTemplatePath;
  FileTemplateCache templateCache = null;

  private TemplateManager() {
    this.init();
    
  }

  public static TemplateManager getManager() {
    if (manager == null)  {
      synchronized(TemplateManager.class)  {
        if (manager == null)  {
          manager = new  TemplateManager();
        }
      }
    }
    return manager;
    }
    
    public void setTemplateRootPath(String rootPath)	{
			rootTemplatePath = rootPath;
    }

  /**
  * init template and cache management.
  * cache template is use to detect cache update during synchronization
  */
  protected void init() {
    String templatesDir = PropertiesManager.getManager().getProperty("org.openmobileis.services.templatesDir");
    rootTemplatePath = org.openmobileis.common.util.file.FileUtilities.convertFileNameToSystem(templatesDir);
    templateCache = new FileTemplateCache(rootTemplatePath);
    templateCache.setFilenameSuffix("htm");
    
  }

  // cache listener methods
  public void cacheUnavailable(CacheEvent e) {
    System.out.println("Cache unavailable: " + e.getException().toString());
  }

  public void elementUpdated(CacheEvent e) {
    System.out.println("Template updated: " + e.getElementName());
  }

  public void elementUpdateFailed(CacheEvent e) {
    System.out.println("Update of template " + e.getElementName() +
            " failed: " + e.getException().toString());
    e.getException().printStackTrace();
  }

  public void elementRemoved(CacheEvent e) {
    System.out.println("Template removed: " + e.getElementName());
  }

  /**
   * initialize template structure for this request.
   */
  public TemplateModelRoot getTemplateModelRoot()  {
    return new SimpleHash();
  }

  /**
   * return the template with specified relative name.
   * The path to find the temlpate is template install directory(intalpath/cn/openmischannel/template/"+relatifName.
   * The relatid name must not begin with file separator.
   * The relative name must is plate-forme specific file separators.
   * Return null if not found.
   */
  public Template getTemplate(String relatifPath) {
    //use template cache to allow template include.
    IntlResourceManager resourceManager = IntlResourceManager.getManager();
    relatifPath = resourceManager.getLocalizedFileName(relatifPath);
    return (Template)templateCache.getTemplate(relatifPath);
 /*   try {
       String fileName = rootTemplatePath+relatifPath;
      fileName = resourceManager.getLocalizedFileName(fileName);
      File testfile = new File(fileName);
      if (testfile.exists())  
        return new Template(fileName);
    } catch (Throwable ex)  {
       LogManager.trace(new ServerException("TemplateNotFound :"+rootTemplatePath+File.separator+relatifPath, ex));
    }
    return null; */
  }

/*
  * Get the template called templateName (load it if it is not already done),
  * replace all variables tree of TemplateModelRoot in the template.
  * Write the HTTP response in the output stream with a basic header (mime type + status)
  * @param: String templateName the name of the template to use
  * @param: Hashtable variables (varName -> variable value) pairs
  * @param: HttpServletResponse response
  * @return: none
  * @OpenMSPException : IOEXception if an error occurs during sending data
  * @OpenMSPException : TemplateNotFoundException if the template is not found
  */
  public void sendResponse (String templateName, TemplateModelRoot modelRoot, HttpServletResponse res) throws java.io.IOException, TemplateNotFoundException {
    if (templateName != null) {
      Template template = this.getTemplate(templateName);
      if (template != null) {
        res.setContentType("text/html" );
        PrintWriter print = new PrintWriter(res.getOutputStream());
        try {
          template.process(modelRoot, print);
        } finally {
          print.flush();
        }
      } else  {
        LogManager.traceError(LogServices.WEBSERVICE, "Template non trouv :"+templateName);
        ServletTools.sendErrorPage("Erreur Interne", "Fichier de visualisation manquant.","/index", res);
      }
    }
  }


}



