package org.openmobileis.modules.profils.embedded.data;

import java.io.IOException;
import java.io.Reader;

import org.kxml2.io.KXmlParser;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;
import org.xmlpull.v1.XmlPullParserException;

public class ProfilParser {
	
	private final static String TAG_XML = "xml";
	private final static String TAG_PROFIL = "profil";
	private final static String TAG_RUBRIC = "rubric";
	private final static String TAG_PROPERTY = "property";
	private final static String TAG_MODULE = "module";
	private final static String TAG_MODRUBRIC = "modrubric";
	
	private Reader reader;
	
	private ProfilRubric currentRubric = null;
	private ProfilModule currentModule = null;
	
	private boolean propertyTagOpened = false;
	private String propertyName = null;
	private String propertyValue = null;
	
	private String modrubricName = null;
	
	private Array rubrics = new Array(5);
	private Array modules = new Array(5);
	
	String currentTag;
	
	public ProfilParser(Reader reader) {
		this.reader = reader;
	}
	
	public void parse() throws XmlPullParserException, IOException {
		KXmlParser parser = new KXmlParser();
		parser.setInput(this.reader);
		
        int eventType = parser.getEventType();
        do {
            if(eventType == KXmlParser.START_DOCUMENT) {
            } else if(eventType == KXmlParser.END_DOCUMENT) {
            } else if(eventType == KXmlParser.START_TAG) {
                this.processStartElement(parser);
            } else if(eventType == KXmlParser.END_TAG) {
                this.processEndElement(parser);
            } else if(eventType == KXmlParser.TEXT) {
                this.processText(parser);
            }
            eventType = parser.next();
        } while (eventType != KXmlParser.END_DOCUMENT);
	}
	
	private void openRubricTag(KXmlParser parser) {
		if (currentRubric != null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, starting a new rubric while the preceding one isn't closed");
		}
		if (currentModule != null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, starting a new rubric while the preceding module isn't closed");
			currentModule = null;
		}
		currentRubric = new ProfilRubric();
        int count = parser.getAttributeCount();
        for (int i = 0; i < count; i++) {
			if (parser.getAttributeName(i).equals("name")) {
				currentRubric.setName(parser.getAttributeValue(i));
			} else {
				LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unrecognized attribute "+parser.getAttributeName(i)+" for rubric");
			}
        }		
	}
	
	private void closeRubricTag() {
		if (currentRubric == null) {
			LogManager.traceError(LogServices.SYNCHROSERVICE, "Corrupted profil file, found rubric closing tag while no rubric is open");
			return;
		}
		rubrics.add(currentRubric);
		currentRubric = null;
	}
	
	private void openModuleTag(KXmlParser parser) {
		if (currentRubric != null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, starting a new module while the preceding rubric isn't closed");
			currentRubric = null;
		}
		if (currentModule != null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, starting a new module while the preceding one isn't closed");
		}
		currentModule = new ProfilModule();
        int count = parser.getAttributeCount();
        for (int i = 0; i < count; i++) {
			if (parser.getAttributeName(i).equals("name")) {
				currentModule.setName(parser.getAttributeValue(i));
			} else if (parser.getAttributeName(i).equals("version")) {
				currentModule.setVersion(parser.getAttributeValue(i));
			} else {
				LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unrecognized attribute "+parser.getAttributeName(i)+" for modrubric");
			}
		}		
	}
	
	private void closeModuleTag() {
		if (currentModule == null) {
			LogManager.traceError(LogServices.SYNCHROSERVICE, "Corrupted profil file, found module closing tag while no module is open");
			return;
		}
		modules.add(currentModule);
		currentModule = null;
	}
	
	private void openPropertyTag(KXmlParser parser) {
		if (currentRubric != null && propertyTagOpened == true) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, opening a new rubric property while the preceding one isn't closed");
		} else if (currentModule != null && propertyTagOpened == true) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, opening a new rubric property while the preceding one isn't closed");
		} else if (currentModule == null && currentRubric == null){
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, argument property must be included in a rubric or in a module");
			return;
		}
		
		propertyTagOpened = true;
		propertyName = null;
		propertyValue = null;

        int count = parser.getAttributeCount();
        for (int i = 0; i < count; i++) {
			if (parser.getAttributeName(i).equals("name")) {
				propertyName = parser.getAttributeValue(i);
			} else if (parser.getAttributeName(i).equals("value")) {
				propertyValue = parser.getAttributeValue(i);
			} else {
				LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unrecognized attribute "+parser.getAttributeName(i)+" for modrubric");
			}
		}
        
        if (propertyName == null) {
        	LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, property tag with no name !");
        }
        if (propertyValue == null) {
        	LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, property tag "+propertyName+" with no value !");
        }
	}
	
	private void closePropertyTag() {
		if (propertyTagOpened == false) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, closing tag property while it's not opened");
		} else if (propertyName == null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, closing tag property while it has no name defined");
		} else if (propertyValue == null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, closing tag property "+propertyName+" while it has no value defined");
		} else {
			if (currentRubric != null) {
				currentRubric.addProperty(propertyName, propertyValue);
			} else if (currentModule != null) {
				currentModule.addProperty(propertyName, propertyValue);
			} else {
				LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, property "+propertyName+" outside rubrics and modules");
			}
		}
		propertyTagOpened = false;
		propertyName = null;
		propertyValue = null;
	}
	
	private void openModrubricTag(KXmlParser parser) {
		if (currentModule == null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, argument modrubric must be included in a module");;
			return;
		}
		
		int count = parser.getAttributeCount();
        for (int i = 0; i < count; i++) {
			if (parser.getAttributeName(i).equals("name")) {
				modrubricName = parser.getAttributeValue(i);
			} else {
				LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unrecognized attribute "+parser.getAttributeName(i)+" for modrubric");
			}
        }
	}
	
	private void closeModrubricTag() {
		if (currentModule == null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, argument modrubric must be included in a module");
			return;
		}
		if (modrubricName == null) {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, argument modrubric must have a name");
			return;
		}
		currentModule.addRubric(modrubricName);
		modrubricName = null;
	}
	
    private void processStartElement (KXmlParser parser)
    {
        String name = parser.getName();
        
        if (name.equals(ProfilParser.TAG_XML)) {
        	// Do nothing
        } else if (name.equals(ProfilParser.TAG_PROFIL)) {
        	// Do nothing
        } else if (name.equals(ProfilParser.TAG_RUBRIC)) {
        	this.openRubricTag(parser);
		} else if (name.equals(ProfilParser.TAG_MODULE)) {
			this.openModuleTag(parser);
		} else if (name.equals(ProfilParser.TAG_PROPERTY)) {
			this.openPropertyTag(parser);
		} else if (name.equals(ProfilParser.TAG_MODRUBRIC)) {
			this.openModrubricTag(parser);
		} else {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unknown opening tag : "+name);
		}
    }


    private void processEndElement (KXmlParser parser)
    {
        String name = parser.getName();
        
        if (name.equals(ProfilParser.TAG_XML)) {
        	// Do nothing
        } else if (name.equals(ProfilParser.TAG_PROFIL)) {
        	// Do nothing
        } else if (name.equals(ProfilParser.TAG_RUBRIC)) {
        	this.closeRubricTag();
		} else if (name.equals(ProfilParser.TAG_MODULE)) {
			this.closeModuleTag();
		} else if (name.equals(ProfilParser.TAG_PROPERTY)) {
			this.closePropertyTag();
		} else if (name.equals(ProfilParser.TAG_MODRUBRIC)) {
			this.closeModrubricTag();
		} else {
			LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unknown closing tag : "+name);
		}
    }

    private void processText (KXmlParser parser) throws XmlPullParserException
    {
    	// Do nothing
    }
	
	public Array getRubrics() {
		return this.rubrics;
	}
	
	public Array getModules() {
		return this.modules;
	}
}
