/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
package org.openmobileis.services.common;

import java.util.Hashtable;

import javax.servlet.http.HttpServletRequest;

import org.openmobileis.common.util.collection.Array;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Calling service manager is use to keep an historic of different service call to be 
 * able to return to a calling service when a specific function has been done.
 * It can be use to profide facility to fill form. For exemple in a form if you want to provide search
 * to fill a property, you register the form calling service and after search the search result service
 * redirect to the registered form service.
 * For each session a calling service is created. Use the service session to get current calling service.
 * Multiple calling service can be registered. A stack structure is use to push and pop service.
 * 
 * The stack can only be peek. Return URL contain a call to the RegisteredServicePopService that pop the URI
 * and call the registered service URI
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class CallingServiceManager {
	
	private Array stack;
	private Array paramList;
	private int stackIndex = 0;
	 
	/**
	 * Constructor for CallingServiceManager.
	 */
	public CallingServiceManager() {
		super();
		stack = new Array(5);
		paramList = new Array(5);
	}

  /** return the stack of registered service.
   * This stack is use to keep an historic of service call to access a function.
   * during service call. Service URI are push on the stack.
   * To return to the last registered service us pop or peek
   *@return : the registered service name. Null if no service is registered
  */
  public Array getRegistedCallingServiceList()	{
  	return stack;
  }
  
  /**
   * return the registered service parameters that is on the top of the stack 
   * return null if no service is registered
   */
  public Hashtable getCallingServiceParameters()	{
  	if (stackIndex > 0)	{
  		return (Hashtable) paramList.get(stackIndex-1);
  	}
  	return null;
  }
  
  /**
   * push the specified service name on the calling service stack
   */
  public void pushCallingService(String serviceURI, HttpServletRequest req)	{
  	stack.add(stackIndex,serviceURI);
  	paramList.add(stackIndex, new Hashtable(req.getParameterMap()));
  	stackIndex++;
  }
  
  /**
   * Removes the calling service URI at the top of this stack and returns that calling service URI as the value of this function.
   * return null if no service is registered
   */
  public String popCallingService() {
  	if (stackIndex > 0)	{
  		String uri = (String)stack.remove(stackIndex-1);
  		paramList.remove(stackIndex-1);
  		stackIndex --;
  		return uri;
  	}
  	return null;
  }
  
  /**
   * Looks at the calling service URI  at the top of this stack without removing it from the stack.
   * return the calling service URI
   * return null if no service is registered
   */
  public String peekCallingService()	{
  	if (stackIndex > 0)	{
  		String uri = (String)stack.get(stackIndex-1);
  		return uri;
  	} 
  	return null;
  }
  
  /**
   * Verify if a service is registered and return the service url that manage calling service redirection.
   * return null if not calling service is registered.
   */
  public String getCallingServiceURL()	{
  	if (this.peekCallingService() != null)	{
  		return ServiceManager.getManager().getServiceBaseURI()+"/admin/callingservice";
  	} else {
  		return null;
  	}
  }
  
  /**
   * Tests if this calling service stack is empty.
   */
  public boolean isEmpty()	{
  	return (stack.size()==0);
  }
  
  /**
   * pop the stack until the specified calling service URI is found. 
   * The service is not removed from the calling 
   * service URI stack.
   * If the service is not found, the stack is not modified.
   * return true if the service URI is found, false otherwise. 
   */
	public boolean peekToTheSpecifiedCallingService(String serviceURI)	{
		Array clonedStack = (Array)stack.clone();
		Object ret;
		for (int i=stackIndex-1; i>=0; i--)	{
			ret = clonedStack.remove(i);
			if((ret !=null)&&(ret.equals(serviceURI)))	{
				stack = clonedStack;
				stackIndex = i+1;
				stack.add(ret);
				return true;
			}
			paramList.remove(i);
			
		}
		return false;
	}
	
	/**
	 * clear the service URI stack
	 */
	public void clearStack()	{
		stack.clear();
		paramList.clear();
		stackIndex = 0;
	}
}
