/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */

package org.openmobileis.services.common;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;
import org.openmobileis.embedded.modules.ModuleLoader;
import org.openmobileis.services.Service;
import org.openmobileis.services.servlet.OpenmisHttpServletRequest;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Load all known services at launching. Names and
 * classes are described in a property file and set at the initialization (init() method) by
 * the web server.
 * Each time an objecthas to access to a service, call getService from Service manager.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class ServiceManager {

  private static ServiceManager manager;
  // key : service name, value : service instance
  private java.util.Hashtable services = new java.util.Hashtable(10);
  private Array modulesList; //keep module list to avoid class garbage
  private String baseURI;// = "/services";


  // init called by the web server at the initialization phase (new Serve())
  public void init (java.util.Properties props, String baseURI) {
        // Init baseURI
        this.baseURI = baseURI;
      
		// load services
		 modulesList = new Array(5);
		 
		 ReturnLinkManagerService returnlinkservice = new ReturnLinkManagerService();
		 String returnlinkname = returnlinkservice.getServiceUri().substring(this.baseURI.length(), returnlinkservice.getServiceUri().length());
		 services.put(returnlinkname, returnlinkservice);
		 
		java.util.Enumeration enume = props.propertyNames();
		while (enume.hasMoreElements()) {
			 String serviceName = (String)enume.nextElement();
			 String className = props.getProperty(serviceName);
				 // Check if we're allowed to make one of these.
			 try {
//	LogManager.traceWarning(0, "SERVICE_MANAGER loading service "+serviceName+" class :"+className);
					 Object toload = Class.forName( className ).newInstance();
					 if (toload instanceof Service) {
						 Service service = (Service)toload;
						 service.setServiceURI(this.baseURI + serviceName);
						 services.put( serviceName, service );
					 } else if (toload instanceof ModuleLoader){
						 ModuleLoader modules = (ModuleLoader)toload;
						 modules.loadService(services);
						 modulesList.add(modules);
					 } else {
					     LogManager.traceError(LogServices.WEBSERVICE, "Try to load a class which is neither a service, neither a module loader : "+className);
					 }
				 } catch ( ClassNotFoundException e )  {
						LogManager.traceError(LogServices.WEBSERVICE, "Class not found: " + className );
						LogManager.traceError(LogServices.WEBSERVICE, e );
				 }	catch ( ClassCastException e )  {
					 LogManager.traceError(LogServices.WEBSERVICE, "Class cast problem: " + e.getMessage());
					LogManager.traceError(LogServices.WEBSERVICE, e );
				 }	catch ( InstantiationException e ) {
					 LogManager.traceError(LogServices.WEBSERVICE, "Instantiation problem: " + e.getMessage() );
					LogManager.traceError(LogServices.WEBSERVICE, e );
				 }	catch ( IllegalAccessException e )  {
					 LogManager.traceError(LogServices.WEBSERVICE, "Illegal class access: " + e.getMessage());
					LogManager.traceError(LogServices.WEBSERVICE, e );
				 }	catch ( Exception e )  {
					 LogManager.traceError(LogServices.WEBSERVICE, "Unexpected problem creating servlet: " + e );
					 LogManager.traceError(LogServices.WEBSERVICE, e );
				} // end catch
			} //end while
  }
  
  public void addModuleLoader(ModuleLoader moduleLoader) {
	  moduleLoader.loadService(services);
	  modulesList.add(moduleLoader);
  }
  
  public void removeModuleLoader(ModuleLoader moduleLoader) {
	  moduleLoader.unloadService(services);
	  modulesList.remove(moduleLoader);
  }
  
  public String getServiceBaseURI()	{
  	return baseURI;
  }
  
	public void setServiceBaseURI(String uri)	{
		baseURI = uri;
	}


  public static ServiceManager getManager() {
  	if (manager == null)	{
  		synchronized(ServiceManager.class)	{
				if (manager == null)	{
					manager = new ServiceManager();
				}
  		}
  	}
    return manager;
  }

  private ServiceManager() {
  }

  /*
  * Return the Service called serviceName. Return null if it is not registered
  * @param : String service name
  * @return: Service or null
  */
  public Service getService (String serviceName) {
    // try to get the service from URL of type /serviceName/otherPath/....
    Service service =  (Service)services.get(serviceName);
    if (service == null)    {
        int index  = serviceName.lastIndexOf('/');
        while (index != -1) {
           serviceName = serviceName.substring(0, index); 
           service =  (Service)services.get(serviceName);
           if (service != null) {
            break;
           }
           index  = serviceName.lastIndexOf('/');
        }
    }
    return service;
  }
  
  public Service getServiceByURI(String serviceURI) {
	  if (serviceURI.startsWith(baseURI)) {
		  return this.getService(serviceURI.substring(baseURI.length()));
	  } else {
		  return this.getService(serviceURI);
	  }
  }

  public boolean exists (String serviceName) {
    return services.containsKey(serviceName);
  }
  
  public void redirectToServiceURI(String uri, HttpServletRequest req, HttpServletResponse res) throws IOException, ServletException {
	  Service service = this.getServiceByURI(uri);
    OpenmisHttpServletRequest request = (OpenmisHttpServletRequest)req;
    if (!uri.startsWith(baseURI)) {
      uri=baseURI+uri;
    } 
    request.setRequestURI(uri);
	  service.runService(req, res);
  }

}