/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */

package org.openmobileis.synchro.openmsp.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.zip.GZIPOutputStream;

import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.context.Plateform;
import org.openmobileis.common.intl.IntlResourceManager;
import org.openmobileis.common.user.UserNotFoundException;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.exception.SynchroException;
import org.openmobileis.embedded.util.ServicePropertiesManager;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.client.SynchroProcessor;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.client.conduit.ConduitParameter;
import org.openmobileis.synchro.openmsp.client.core.ListenerOrdonancer;
import org.openmobileis.synchro.openmsp.client.core.OpenMSPSyncMessageFactory;
import org.openmobileis.synchro.openmsp.protocol.Message;
import org.openmobileis.synchro.openmsp.protocol.MessageFactory;
import org.openmobileis.synchro.openmsp.protocol.Status;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Default SynchroManager for OpenMSP protocol.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class OpenMSPSynchroManager implements SynchroProcessor {
  public static final String OPENMSPPROCESSORNAME="OpenMSPSync";

  private static OpenMSPSynchroManager manager;
  private ListenerOrdonancer ordonancer;
  private OpenMSPSyncMessageFactory messageFactory;
  private String lastSynchroLogin;
  private int globalSynchroStatut;

  /**
   * 
   */
  private OpenMSPSynchroManager() {
    super();
    this.ordonancer = new ListenerOrdonancer();
    this.messageFactory = new OpenMSPSyncMessageFactory();
  }

  public static OpenMSPSynchroManager getManager() {
    if (manager == null)  {
      synchronized (OpenMSPSynchroManager.class) {
        if (manager == null)  {
          manager = new OpenMSPSynchroManager();
          ApplicationContextManager.getManager().addManager(manager);
        }
      }
    }
    return manager;
  }
 
  public String getProcessorName()  {
    return OpenMSPSynchroManager.OPENMSPPROCESSORNAME;   
  }

  public void notifyFailure(String listernerName) {
    OpenMSPSyncListener listener = this.getListenerByName(listernerName);
    if (listener != null) {
      listener.notifySynchroFailure();
      ordonancer.notifySynchroError(listener.getSyncName());
    }
    globalSynchroStatut = Status.STATUS_FAILED;
  }
  
  public OpenMSPSyncListener getListenerByName(String listenerName) {
    return ordonancer.getListenerByName(listenerName);
  }
  
  public void doSynchro(Credential cred, SynchroDescriptor descriptor) throws SynchroException {
    globalSynchroStatut = Status.STATUS_SYNC_IN_PROGRESS;
    this.lastSynchroLogin = cred.getPrincipal();
    //save login
    ServicePropertiesManager.getManager().saveProperty("OpenMSPSynchroManager", "login", this.lastSynchroLogin);
    this.ordonancer.beginSynchro();
    this.messageFactory.beginSynchro(cred, descriptor);
    Array listenerList = this.ordonancer.getNextListenerList();
    try {
      while (listenerList.size() != 0)  {
        int listsize = listenerList.size();
        for (int i=0; i<listsize; i++)  {
          OpenMSPSyncListener listener = (OpenMSPSyncListener) listenerList.get(i);
          listener.startSync(cred, descriptor);
        }
        while (true) {
          Message message = this.messageFactory.getOpenMSPMessage(listenerList);
          if (message.isEmpty()) {
            //get next listener phase.
            break;
          }
          String xmlString = message.encode();
          ConduitParameter[] parameters = new ConduitParameter[] {new ConduitParameter("openMISGroup", descriptor.getSynchroGroup())};
 
          String reqResult = descriptor.getSynchroConduit().sendRequest(parameters, new String(compressAndEncodeRequest(xmlString)), descriptor.getServerURL());
          // third step : Process responses
          Message receivedMessage = MessageFactory.getFactory().getMessage(reqResult);
          String target = receivedMessage.getHeader().getTarget();
          if (!target.equals(OpenMSPSyncMessageFactory.OpenMSPSOURCE)) {
            throw new OpenMSPException("Bad OpenMSP Sync Source in Server answer message. Synchro Abort.");
          } else {
            this.messageFactory.receiveOpenMSPMessage(listenerList, receivedMessage);
          }
        }
        for (int i=0; i<listsize; i++)  {
          OpenMSPSyncListener listener = (OpenMSPSyncListener) listenerList.get(i);
          listener.endSync();
        }
        listenerList = this.ordonancer.getNextListenerList();
      }
      globalSynchroStatut = Status.STATUS_OK;
    } catch (UserNotFoundException ex)  {
      notifyFailureToListeners(listenerList);
      globalSynchroStatut = Status.STATUS_UNAUTHORIZED;
      throw new OpenMSPException(ex);
    } catch (Throwable ex)  {
      notifyFailureToListeners(listenerList);
      throw new OpenMSPException(ex);
    } finally {
      this.ordonancer.endSynchro();
      this.messageFactory.endSynchro();      
    }
  }
  
  /**
   * register a new listener and its depends that will participate to the synchro.
   * Listener's depends are an array of listener name that must be synchronized before
   * the new listener.
   * Thier is no specific order to register the listener and its dependance.
   * @param listener : listener to register
   * @param depends : list of listener name 
   */
  public void addListener(OpenMSPSyncListener listener, String[] depends) {
    ordonancer.addListener(listener, depends);
  }
  
  private void notifyFailureToListeners(Array listenerList) {
    for (int i=0; i<listenerList.size(); i++) {
      OpenMSPSyncListener listener = (OpenMSPSyncListener)listenerList.get(i);
      listener.notifySynchroFailure();
      ordonancer.notifySynchroError(listener.getSyncName());
    }
    globalSynchroStatut = Status.STATUS_FAILED;   
  }

  private byte[] compressAndEncodeRequest(String request) throws IOException {
    // compresse data
    byte[] compressedData = null;
    ByteArrayOutputStream byteStream = new ByteArrayOutputStream(512);
    GZIPOutputStream zipStream = new GZIPOutputStream(byteStream);
    try {
      zipStream.write(request.getBytes());
    } finally  {
      zipStream.close();
    }
    compressedData = byteStream.toByteArray();

    // encode compressed data
    byte[] returnData = GeneralCoder.encodeBase64(compressedData);
    return returnData;
  }

  public String getLastSynchroLogin() {
    if (this.lastSynchroLogin == null){
      this.lastSynchroLogin = ServicePropertiesManager.getManager().getProperty("OpenMSPSynchroManager", "login");

    }
    return this.lastSynchroLogin;
  }
  
  /**
   * return the user agent to use with synchronisation.
   * The user agent can be defined with the property org.openmobileis.synchro.direct.useragent.
   * or if not set, it's constructed with plateform information.
   * default user agent : OpenMSPPlug-<OpenMobileISMajorVersion>.<OpenMobileISMinorVersion>-<OS>[<OSVersion>]-<user language>-<synchro type if present>
   * The synchro type is defined with the property OpenMSPsynchrotype of the descriptor.
   * @return
   */
  public String getSynchroUserAgent(SynchroDescriptor descriptor) {
    String usera = PropertiesManager.getManager().getProperty("org.openmobileis.synchro.direct.useragent");      
    if (usera == null)  {
      Plateform plateform = ApplicationContextManager.getManager().getApplicationContext().getPlateform();
      StringBuffer useragent = new StringBuffer("OpenMSPPlug-");
      useragent.append(plateform.getOpenMobileISMajorVersion()).append('.').append(plateform.getOpenMobileISMinorVersion()).append('-');
      useragent.append(plateform.getOS()).append('[').append(plateform.getOSVersion()).append("]-");
      String language = IntlResourceManager.getManager().getManagerLocaleString();
      useragent.append(language);
      String synctype = descriptor.getProperty("OpenMSPsynchrotype");
      if (synctype != null) {
        useragent.append('-').append(synctype);
      }
      usera = useragent.toString();
    }
    return usera;
  }

  public int getGlobalSynchroStatut() {
    return globalSynchroStatut;
  }

}
