/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.synchro.openmsp.client.core;

import org.openmobileis.common.util.collection.Array;
import org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Manage listener order during different synchro phase.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class ListenerOrdonancer {
  private ListenerPhaseList firstPhase, currentPhase;
  private static final Array emptyArray = new Array();

  /**
   * 
   */
  public ListenerOrdonancer() {
    super();
  }
  
  public OpenMSPSyncListener getListenerByName(String listenerName) {
    ListenerPhaseList phase = firstPhase;
    while (phase != null) {
      ListenerNode node = this.getNodeFromPhaseList(listenerName);
      if (node != null) {
        return node.listener;
      }
      phase = phase.nextPhase;
    }
    return null;
  }
  
  public void addListener(OpenMSPSyncListener listener, String[] dependList) {
    if (firstPhase == null)  {
      firstPhase = new ListenerPhaseList(1);
    }
    Array depends = null;
    if (dependList == null) {
      depends = emptyArray;
    } else  {
     depends = new Array(dependList);
    }
    //first add listener dependance.
    Array depArray = new Array();
    int dependssize = depends.size();
    for (int i=0; i<dependssize; i++)  {
     ListenerNode depnode = new ListenerNode();
     depnode.listenerName = (String)depends.get(i);
     this.addListenerToList(depArray, firstPhase, depnode); 
    }
    //add new listener
    ListenerNode newnode = new ListenerNode();
    newnode.listenerName = listener.getSyncName();
    newnode.listener = listener;
    
    //process depends will not all treated. Always end because all depends has already been added.
    do  {
      depends = this.addListenerToList(depends, firstPhase, newnode);  
    } while (depends.size() != 0);
  }
  
  public void beginSynchro()  {
    //set all synchro listener node to OK
    ListenerPhaseList phase = firstPhase;
    while (phase != null) {
      int size = phase.nodeList.size();
      for (int i=0; i<size; i++)  {
        ListenerNode node = (ListenerNode) phase.nodeList.get(i);
        node.validState = true;
      }
      phase = phase.nextPhase;
    }
    currentPhase = firstPhase;
  }
  
  public void endSynchro()  {
  }
  
  public Array getNextListenerList()  {
    if (currentPhase == null) return ListenerOrdonancer.emptyArray;
    Array listenerList = new Array();
    //remove not ok listener
    int size = currentPhase.nodeList.size();
    for (int i=0; i<size; i++)  {
      ListenerNode node = (ListenerNode) currentPhase.nodeList.get(i);
      if (node.validState)  {
        listenerList.add(node.listener);
      }
    }
    currentPhase = currentPhase.nextPhase;
    return listenerList;
 }
  
  public void notifySynchroError(String listenerName) {
    ListenerNode listener = this.getNodeFromPhaseList(listenerName);
    if (listener != null) {
      invalidListenerChilds(listener);
    }
  }
  
  private void invalidListenerChilds(ListenerNode listener)  {
    listener.validState = false;
    int size = listener.nexts.size();
    for (int i=0; i<size; i++)  {
      this.invalidListenerChilds((ListenerNode)listener.nexts.get(i));
    }
  }
  
  private ListenerNode getNodeFromPhaseList(String listenerName)  {
    ListenerPhaseList phase = firstPhase;
    while (phase != null) {
      int size = phase.nodeList.size();
      for (int i=0; i<size; i++)  {
        ListenerNode node = (ListenerNode) phase.nodeList.get(i);
        if (node.listenerName.equals(listenerName)) {
          return node;
        }
      }
      phase = phase.nextPhase;
    }
    return null;
  }
  
  private Array addListenerToList(Array depends, ListenerPhaseList phase, ListenerNode newnode) {
          
    int phasesize = phase.nodeList.size();
    ListenerNode phaseNode = null;
   
   //test if node already present process already added node
    for (int j=0; j<phasesize; j++)  {
      phaseNode = (ListenerNode)phase.nodeList.get(j);
      if (newnode.listenerName.equals(phaseNode.listenerName)) {
        if (newnode.listener != null) {
          phaseNode.listener = newnode.listener;
        }
        // node already in a phase, use already added node to keep node context data.
         newnode = phaseNode;
      }    
    }
    
    //if no dependence add listener
    if (depends.size() == 0)  {
      //validate if not already added in posterior phase.
      if ((newnode.nodePhase != null) && (newnode.nodePhase.phaseNumber < phase.phaseNumber)) {
        newnode.nodePhase.nodeList.remove(newnode);
        newnode.nodePhase = null;
      }
      //if not in a phase add it. //we suppose that all dependance has been added.
      if (newnode.nodePhase == null)  {
        //only add if not already added. Avoid multiple depend add.
        if(this.getNodeFromPhaseList(newnode.listenerName)== null)  {
          newnode.nodePhase = phase;
          phase.nodeList.add(newnode);
          if (newnode.nexts.size() >0)  {
            //move all dependant listerner to next phase and dependent dependant listener and so on.
            movedependentListenerTonextPhase(newnode);
          }
        }
      }
      return ListenerOrdonancer.emptyArray;
    }   
    
    // validate dependance if depend listener is already registered
   Array newdepends = (Array)depends.clone();
   int dependssize = depends.size();
   
   Array clonePhaseNodeList = (Array) phase.nodeList.clone();
   for (int i=0; i<dependssize; i++)  {
      for (int j=0; j<phasesize; j++)  {
        phaseNode = (ListenerNode)clonePhaseNodeList.get(j);
        // test if depends are in this phase.
        if (depends.get(i).equals(phaseNode.listenerName)) {
          newdepends.remove(phaseNode.listenerName);
          phaseNode.nexts.add(newnode);
          if (phase.nextPhase == null)  {
            phase.nextPhase = new ListenerPhaseList(phase.phaseNumber+1);
          }
          newdepends = this.addListenerToList(newdepends, phase.nextPhase, newnode);
        }
      }
      if ((phase.nextPhase != null) && (depends.size() >0))  { //perhaps depends are in next phase.
        newdepends = this.addListenerToList(newdepends, phase.nextPhase, newnode);
      }
    }    
    return newdepends;
  }

  private void movedependentListenerTonextPhase(ListenerNode node)  {
    Array othernodelist = node.nexts;
    node.nexts = new Array();
    Array newdepends = new Array();
    newdepends.add(node.listenerName);
    for (int i=0; i<othernodelist.size(); i++)  {
      ListenerNode listener = (ListenerNode) othernodelist.get(i);
      
      //PB if depend are in next Phase not added.
      this.addListenerToList(newdepends, listener.nodePhase, listener);
      this.movedependentListenerTonextPhase(listener);
    }
    
  }
}


