/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */

package org.openmobileis.synchro.openmsp.client.db;

import java.io.*;

import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.synchro.openmsp.OpenMSPException;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Utility methode to encode and decode rows for the synchro protocol
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class DBImportFileCoder {
  public static final char ADDAction ='A';
  public static final char REPLACEAction= 'R';
  public static final char DELETEAction= 'D';
  public static final char endLine = '\n';
  public static final char dataSeparator = ';';
  public static final char stringSeparator = '\'';
	public static final char antislashSeparator = '\\';
	public char guillemetannulator = '\\';
	private static DBImportFileCoder coder;
	
	private DBImportFileCoder()	{
		String newGuillemetAnnulator = PropertiesManager.getManager().getProperty("server.common.util.string.format.sql.guillement");
		if (newGuillemetAnnulator != null)	{
			guillemetannulator = newGuillemetAnnulator.charAt(0);
		}
	}
	
	public static DBImportFileCoder getCoder()	{
		if (coder==null)	{
			synchronized(DBImportFileCoder.class)	{
				if (coder == null)	{
					coder = new DBImportFileCoder();
				}
			}
		}
		return coder;
	}

  /**
   * encode the specified row data and action in a char array that can be decoded.
   * String row data must be well formated with cote delimiter :ex tutu => 'tutu'
   */
  public char[] encodeData(char action, String[] rowData) throws OpenMSPException {
    CharArrayWriter writer = new CharArrayWriter();
    try {
      writer.write(action);
      writer.write(dataSeparator);
      for (int i=0; i<rowData.length; i++) {
        if (rowData[i] == null) {
          writer.write("NULL");
        } else  {
          writer.write(rowData[i]);
        }
        if (i!=rowData.length -1)  {
          writer.write(dataSeparator);
        }
      }
      writer.write(endLine);
    } catch (Exception ex)  {
      throw new OpenMSPException("DBImportFileCoder encodeData Error ", ex);
    } finally {
      writer.flush();
      writer.close();
    }
    return writer.toCharArray();
  }

  public String formateString(String data) {
    StringBuffer str = new StringBuffer(stringSeparator);
    str.append(data).append(stringSeparator);
    return str.toString();
  }

  public String removeAntiSlash(String input)  {
    char[] chs = input.toCharArray();
    int size = chs.length;
    CharArrayWriter out = new CharArrayWriter(30);
    for (int i=0; i < size; i++)  {
      if (chs[i] == antislashSeparator) {
        continue;
      }
      out.write(chs[i]);
    }
    out.flush();
    out.close();
    return out.toString();
  }


  /**
   * decode the specified file and notify the specified listener when a row is decoded.
   */
  public void decodeFileData(String filename, DecodedRowListener listener) throws OpenMSPException {
    try {
      BufferedReader reader = new BufferedReader(new FileReader(filename));
      try {
        char read = 0;
        int input = 0;
        boolean insideString = false;
        StringBuffer str = new StringBuffer();
        Array dataList = new Array(10);
        boolean seeStringSeparator = false;
        boolean seeAntiSeparator = false;
        int line =0;
        char action =0;
        boolean rowBeginned = false;
        while ((input = reader.read()) != -1) {
          read=(char) input;
          if (!insideString)  {
            if (read == dataSeparator)  {
              if (rowBeginned)  {
                String toAdd = str.toString();
                if (toAdd.equals("NULL"))  {
                  dataList.add(null);
                } else  {
                  dataList.add(toAdd);
                }
                str.setLength(0);
              } else  {
                action = str.charAt(0);
                rowBeginned = true;
                str.setLength(0);
              }
              continue;
            } else if (read == stringSeparator)  {
              insideString = true;
              seeStringSeparator = false;
              seeAntiSeparator = false;
              continue;
            } else if (read == endLine) {
              line++;
              String toAdd = str.toString();
              if (toAdd.equals("NULL"))  {
                dataList.add(null);
              } else  {
                dataList.add(toAdd);
              }
              str.setLength(0);
              // notify row.
              String[] rowData = new String[dataList.size()];
              dataList.toArray(rowData);
              listener.notifyNewRowDecoded(action, rowData);
              dataList.clear();
              rowBeginned = false;
              continue;
            }
          } else  {
            if (read == antislashSeparator)  {
                seeAntiSeparator = true;
            } else if (read == stringSeparator)  {
              if (seeAntiSeparator) {
                seeAntiSeparator = false;
              } else  {
                seeStringSeparator = true;
                continue;
              }
            } else if (seeStringSeparator) {
              insideString = false;
              if (read == dataSeparator)  {
                dataList.add(str.toString());
                str.setLength(0);
              } else if (read == endLine) {
                line++;
                dataList.add(str.toString());
                str.setLength(0);
                // notify row.
                String[] rowData = new String[dataList.size()];
                dataList.toArray(rowData);
                listener.notifyNewRowDecoded(action, rowData);
                dataList.clear();
                rowBeginned = false;
                continue;
              } else  {
                listener.notifyError(line, read);
              }
              continue;
            } else  if (seeAntiSeparator){
              seeAntiSeparator = false;
            }
          }
          str.append(read);
        }
      } finally {
        reader.close();
      }
    } catch (Exception ex) {
      throw new OpenMSPException("DBImportFileCoder decodeFileData Error ", ex);
    }
  }

  public String cleanAndFormatStringforDB(String toFormat)  {
    if (toFormat == null)
      return "NULL";
    return new String(cleanAndFormatStringforDB(toFormat.toCharArray()));
  }

  private char[] cleanAndFormatStringforDB(char[] toFormat)  {
  	
    CharArrayWriter writer  = new CharArrayWriter();
    for (int i=0; i<toFormat.length; i++) {
			if (toFormat[i] == '\'')  {
				writer.write(guillemetannulator);
			} else if (toFormat[i] == antislashSeparator)  {
				writer.write(antislashSeparator);
			} else if (toFormat[i] == 0)  {
				writer.write(' ');
				continue; //skip caractere 00
      }
      writer.write(toFormat[i]);
    }
    writer.flush();
    writer.close();
    return writer.toCharArray();
  }
  
  public String decodeString(String inputString) {
    if (inputString == null) {
      return ("NULL");
    }
    return DBImportFileCoder.getCoder().removeAntiSlash(inputString);
  }

  public String encodeInString(String inputString)  {
    if (inputString == null) {
      return ("NULL");
    }
    char[] toEncode = cleanAndFormatStringforDB(inputString.toCharArray());
    char[] encoded = new char[toEncode.length+2];
    encoded[0] = '\'';
    encoded[toEncode.length+1] = '\'';
    System.arraycopy(toEncode, 0, encoded, 1, toEncode.length);
    return new String(encoded);
  }

  /**
   * decode the specified data into a String array. The specified String is the string encoded by the encode method.
   * The action specified in the string is not return.
   * if an error occurs return null;
   */
  public String[] decodeData(String data)  {
    char read = 0;
    boolean insideString = false;
    StringBuffer str = new StringBuffer();
    Array dataList = new Array(10);
    boolean seeStringSeparator = false;
    boolean seeAntiSeparator = false;
    for (int i=2; i<data.length(); i++) { // begin after the action
      read=data.charAt(i);
      if (!insideString)  {
        if (read == dataSeparator)  {
          String toAdd = str.toString();
          if (toAdd.equals("NULL"))  {
            dataList.add(null);
          } else  {
            dataList.add(toAdd);
          }
          str.setLength(0);
          continue;
        } else if (read == stringSeparator)  {
          insideString = true;
          seeStringSeparator = false;
          seeAntiSeparator = false;
          continue;
        } else if (read == endLine) {
          break;
        }
      } else  {
        if (read == antislashSeparator)  {
          seeAntiSeparator = true;
        } else if (read == stringSeparator)  {
          if (seeAntiSeparator) {
            seeAntiSeparator = false;
          } else  {
            seeStringSeparator = true;
            continue;
          }
        } else if (seeStringSeparator) {
          insideString = false;
          if (read == dataSeparator)  {
            dataList.add(str.toString());
            str.setLength(0);
          } else if (read == endLine) {
            break;
          } else  {
            return null;
          }
          continue;
        } else  if (seeAntiSeparator){
          seeAntiSeparator = false;
        }
      }
      str.append(read);
    }
    

    // add the current buffer
    String toAdd = str.toString();
    if (toAdd.equals("NULL"))  {
      dataList.add(null);
    } else  {
      dataList.add(toAdd);
    }
    String[] rowData = new String[dataList.size()];
    dataList.toArray(rowData);
    return rowData;
  }
  
  public String convertBooleanToDBData(boolean bool)	{
  	if (bool) return "1";
  	return "0";
  }
  
  public boolean convertDBDataToBoolean(byte data)	{
  	if (data == 0) return false;
  	return true;
  }

  public String[] serializeDBObject (String uid, Object obj) throws IOException {
    String[] data = new String[2];
    data[0] = uid;
    ByteArrayOutputStream bstr = new ByteArrayOutputStream();
    ObjectOutputStream ostr = new ObjectOutputStream(bstr);
    ostr.writeObject(obj);
    byte[] b = bstr.toByteArray();
    data[1] = new String(GeneralCoder.encodeBase64(b));
    return data;
  }
  
  public Object unserializeDBObject(String[] synchroData) throws IOException, ClassNotFoundException {
      ByteArrayInputStream bstr = new ByteArrayInputStream(GeneralCoder.decodeBase64(synchroData[1].getBytes()));
      ObjectInputStream ostr = new ObjectInputStream(bstr);
      return ostr.readObject();
  }

  public static void main(String[] args) {
    try	{
	    DecodedRowListener listener = new DecodedRowListener()  {
	      public void notifyNewRowDecoded(char action, String[] rowData) throws OpenMSPException {
	        String str = new String(DBImportFileCoder.getCoder().encodeData(action, rowData));
	        System.out.println("Decoded row :"+str);
	      }
	
	      public void notifyError(int ligne, char errorCharacter) {
	        System.out.println("Error found line "+ligne+" char "+errorCharacter);
	      }
	    };
	
	    String[] addData = {"1234", "'first String'", "'test inside \\'\\'inside\\'\\' fin'", "345"};
	    String[] deleteData = {"'second  String line end \\n suite \\n'", "23", "'test inside \\'\\'inside\\'\\''"};
	    String[] updateData = {"'thrid  separator ;;\\'\\';\\'\\''","'String'","''",null, "345"};
	
	    System.out.println("try file encoding");
	    try {
	      FileWriter writer = new FileWriter("/home/pdelrieu/dev/temp/testimportfile.txt");
	      char[] endoded = DBImportFileCoder.getCoder().encodeData(DBImportFileCoder.ADDAction, addData);
	      System.out.println("Endoded :"+new String(endoded));
	      writer.write(endoded);
	
	      endoded = DBImportFileCoder.getCoder().encodeData(DBImportFileCoder.DELETEAction, deleteData);
	      System.out.println("Endoded :"+new String(endoded));
	      writer.write(endoded);
	
	      endoded = DBImportFileCoder.getCoder().encodeData(DBImportFileCoder.REPLACEAction, updateData);
	      System.out.println("Endoded :"+new String(endoded));
	      writer.write(endoded);
	
	      writer.flush();
	      writer.close();
	
	      DBImportFileCoder.getCoder().decodeFileData("/home/pdelrieu/dev/temp/testimportfile.txt", listener);
	    } catch (Exception ex)  {
	      ex.printStackTrace();
	    }
	
	    String encoded = new String(DBImportFileCoder.getCoder().encodeData(DBImportFileCoder.ADDAction, addData));
	    String[] data = DBImportFileCoder.getCoder().decodeData(encoded);
	    StringBuffer buff = new StringBuffer("Decoded Data :");
	    for (int i=0; i<data.length; i++) {
	      buff.append('#').append(data[i]);
	    }
	    System.out.println(buff.toString());
	
	    encoded = new String(DBImportFileCoder.getCoder().encodeData(DBImportFileCoder.DELETEAction, deleteData));
	    data = DBImportFileCoder.getCoder().decodeData(encoded);
	    buff = new StringBuffer("Decoded Data :");
	    for (int i=0; i<data.length; i++) {
	      buff.append('#').append(data[i]);
	    }
	    System.out.println(buff.toString());
	
	
	    encoded = new String(DBImportFileCoder.getCoder().encodeData(DBImportFileCoder.REPLACEAction, updateData));
	    data = DBImportFileCoder.getCoder().decodeData(encoded);
	    buff = new StringBuffer("Decoded Data :");
	    for (int i=0; i<data.length; i++) {
	      buff.append('#').append(data[i]);
	    }
	    System.out.println(buff.toString());
    } catch (Exception ex)	{
      ex.printStackTrace();
    }


  }
}