/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */

package org.openmobileis.synchro.openmsp.protocol;

import org.openmobileis.synchro.openmsp.OpenMSPException;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Represents a OpenMSP message container a Header and set of commands
 * A Message has a counter to manage cmdId and a ContainerMessage representing the root element.
 * A ContainerMessage contains a Element as value and an array for the children (which are either ContainerMessage or AtomicMessage).
 * To create a command, you have to know if it is a container or an atomic command element, BUT YOU NEVER CREATE AtomicMessage, you've just
 * add the Element to the parent.
 * To create a new message or a message from a string containing an XML representation, call
 * MessageFactory.getFactory() which in charge of creating messages.
 * **********************************************
 * **** How to build a OpenMSP message ? ********
 * **********************************************
 *  1. Get a Message from the factory : MessageFactory.getFactory().createMessage(new Header(sessionID, source, target))
 * you can also create an header by giving credential information.
 *  2. Add commands to the Message.
 *  Message.add(aContainerMessage) if the commands contains subcommands
 *  or Message.add(aElement) if the element is an atomic command such as Status or Item.
 *  3. Create a request command having source and target (Sync or Map)
 *  new CmlContainerCommand(new Command(message.nextCmdId(), Element.SYNC, aSource, aTarget))
 *  4. Create a simple request command : Get, Add, Delete, Replace
 *  new CmlContainerCommand(new CmlResquestCommand(message.nextCmdId(), Element.ADD))
 *  5. Add items to a command
 *  myMessageContainer.add(new Item(Element.ITEM, source, target)) or
 *  myMessageContainer.add(new DataItem(Element.ADD, data, meta, source, target))
 *  myMessageMapContainer.add(new Item(Element.MAP_ITEM, source, target))
 *  6. Create a result
 *  new CmlContainer(new Result(message.nextCmdId(), cmdRef, meta, sourceRef, targetRef))
 *  7. Create a status
 *  new CmlContainer(new Status(message.nextCmdId(), cmdRef, Status.STATUS_OK)
 * ******************************************************
 * See org.openmobileis.common.synchro.OpenMSP.tests.Example to have a complete example
 *
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */

public class Message  implements java.io.Serializable  {
  /**
   * Set version of the serialized data to make it compatible with
   * new class versions.
  */
  final static long serialVersionUID = 1L;


  protected ContainerMessage root;
  protected int cmdIdCounter = 0;
  protected boolean isFinal = true;


  protected Message(Header header, boolean isFinal) {
    root = new ContainerMessage(header);
    this.isFinal = isFinal;
  }

  protected Message (ContainerMessage root, boolean isFinal) {
    this.root = root;
     this.isFinal = isFinal;
  }

  protected void setFinal (boolean isFinalMessage) {
    isFinal = isFinalMessage;
  }

  public boolean isFinal() {
    return isFinal;
  }

  /* return true is there is no command */
  public boolean isEmpty() {
    return root.getMessagesSize() == 0;
  }

  public void add (ContainerMessage container) throws OpenMSPException {
    this.setCmdIds(container);
    root.add(container);
  }

  public void add (Element element)  throws OpenMSPException {
    if ((element.hasCommmandId()) && (element.getCmdId()==0))
      element.setCmdId(this.nextCmdId());
    root.add(element);
  }

  public ContainerMessage getRoot() {
    return root;
  }

  public boolean hasMoreMessage() {
    return root.hasMoreMessage();
  }

  public void resetCursor() {
    root.resetCursor();
  }

  public ContainerMessage nextMessage() {
    return root.nextMessage();
  }

  protected int nextCmdId() {
    cmdIdCounter++;
    return cmdIdCounter;
  }

  public Header getHeader () {
    return (Header)root.getElement();
  }


  public String encode() {
    StringBuffer buffer = new StringBuffer();
    Header header = (Header)root.getElement();
    MessageFactory.writeBeginTag(buffer, MessageFactory.TAG_ROOT);
    buffer.append("\n");
    header.writeBeginTag(buffer);
    header.writeContent(buffer);
    header.writeEndTag(buffer);
    MessageFactory.writeBeginTag(buffer, MessageFactory.TAG_BODY);
    buffer.append("\n");
    for (int i=0; i < root.children.size(); i++) {
      ((ContainerMessage)root.children.get(i)).encode(buffer);
    }
    if (isFinal)
        MessageFactory.writeSingleTag(buffer, MessageFactory.TAG_FINAL);
    MessageFactory.writeEndTag(buffer, MessageFactory.TAG_BODY);
    MessageFactory.writeEndTag(buffer, MessageFactory.TAG_ROOT);
    return buffer.toString();
  }

 protected void setCmdIds (ContainerMessage container) {
    container.resetCursor();
    if ((container.getElement().hasCommmandId()) && (container.getElement().getCmdId()==0))
      container.getElement().setCmdId(this.nextCmdId());
    while (container.hasMoreMessage()) {
          setCmdIds(container.nextMessage());
    }
 }


}

