/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2005 Philippe Delrieu
 * All rights reserved.
 * Contact: openmobileis@e-care.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */

package org.openmobileis.synchro.openmsp.server;


import org.openmobileis.common.user.User;
import org.openmobileis.common.user.UserManager;
import org.openmobileis.common.user.UserNotFoundException;
import org.openmobileis.common.user.profil.Profil;
import org.openmobileis.common.user.profil.ProfilDataManager;
import org.openmobileis.common.user.profil.ProfilNotFoundException;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.common.context.*;
import org.openmobileis.common.util.exception.*;
import org.openmobileis.common.util.log.*;
import org.openmobileis.services.Service;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.client.core.OpenMSPSyncMessageFactory;
import org.openmobileis.synchro.openmsp.protocol.AbstractCommand;
import org.openmobileis.synchro.openmsp.protocol.Command;
import org.openmobileis.synchro.openmsp.protocol.ContainerMessage;
import org.openmobileis.synchro.openmsp.protocol.Element;
import org.openmobileis.synchro.openmsp.protocol.Header;
import org.openmobileis.synchro.openmsp.protocol.Message;
import org.openmobileis.synchro.openmsp.protocol.MessageFactory;
import org.openmobileis.synchro.openmsp.protocol.RequestCommand;
import org.openmobileis.synchro.openmsp.protocol.Result;
import org.openmobileis.synchro.openmsp.protocol.Status;
import org.openmobileis.synchro.openmsp.server.synctarget.SynchroTargerManager;
import org.openmobileis.synchro.openmsp.server.synctarget.proxy.ProxySyncroTarget;
import org.openmobileis.synchro.openmsp.server.util.FileSystem;
import org.openmobileis.synchro.openmsp.server.util.FileSystemObjectPool;
import org.openmobileis.synchro.openmsp.server.util.FileSystemObjectPoolSelecter;
import org.openmobileis.synchro.openmsp.server.util.MemoryFile;
import org.openmobileis.synchro.openmsp.server.util.SynchroFileGenerator;
import org.openmobileis.synchro.openmsp.server.util.SynchroFileGeneratorManager;
import org.openmobileis.synchro.security.auth.Credential;
import org.openmobileis.synchro.security.auth.CredentialCodec;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.zip.*;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: OpenMLservice that implements Nomad-Link synchronisation.
 * @author  Philippe Delrieu
 * @since JDK 1.1
 * @version 1.0.
 */
public class OpenMSPService extends Service{

  public OpenMSPService() {
    String synchrologfile =  PropertiesManager.getManager().getProperty("org.openmobileis.synchro.openmsp.server.log.filename");
    if (synchrologfile != null)  {
      java.util.Properties props = new java.util.Properties();
      props.put("LOGFILE", synchrologfile);
      OpenMSPServiceTracer tracer = new OpenMSPServiceTracer();
      tracer.setLogTracerProperties(props);
      LogManager.attachLogTracerToPriority(LogPriorities.USERSYNCHRO, tracer);
    }
  }
  
  public void run(HttpServletRequest request, HttpServletResponse res) throws ServletException, IOException  {
    // validate client
    User user = null;
    String group = request.getParameter("openMISGroup");
    if (group == null)  {
      LogManager.traceError(0, "OpenMSPService bad request no openMISGroup parameter.");
      res.setStatus(HttpServletResponse.SC_FORBIDDEN);
      return;
    }



    try {
      // get cyberML message
      String openMSPString = request.getParameter("openMISData");
      try {
        openMSPString = new String(uncompressAndDecodeRequest(openMSPString.getBytes()));
      } catch (Throwable ex)  {
        // not encoded or compressed data.
        LogManager.traceError(0, ex);
        res.setStatus(HttpServletResponse.SC_BAD_REQUEST);
        return;
      }


      Message openMLMessage =  MessageFactory.getFactory().getMessage(openMSPString);
      
      Header header = openMLMessage.getHeader();

      // validate user auth.
      String[] credential = header.getCredential();
      Credential cred = null;
      if (credential != null) {
        cred = CredentialCodec.getOpenMSPCredential(credential);
        String userId = UserManager.getManager().authenticateUser(group, cred.getPrincipal(), cred.getPassword());
        user = UserManager.getManager().getUser(group,userId);
      } else  {
        LogManager.traceError(0, "OpenMSPService no credential in OpenMSP request.");
        res.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
        return;
      }
      Profil profil = ProfilDataManager.getManager().getProfil(group, user.getProfil());
      SessionContext context = SessionContextManager.getManager().getSessionContext();
      UserTerminal terminal = (UserTerminal)context.getTerminal();
      if (terminal == null) {
        String useragent = UserTerminal.getUserAgentFromRequest(request);
        terminal = new UserTerminal(useragent);
        context.setTerminal(terminal);
      } 
      
      if (!terminal.isOPenMISClient())  {
        LogManager.traceError(0, "Synchro terminal is not supported for user:"+user.getLogin());
        res.setStatus(HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE);
        return;
      }
      context.setUserId(user.getId());
      context.setUserGroup(group);

      // log new synchro
      LogManager.traceUserSynchro(LogServices.SYNCHROSERVICE, request.getRemoteAddr()+"#"+user.getId()+"#"+group+"#"+terminal.getClientPlateformType()+"#"+terminal.getSynchroType());

      long sessionNumber = header.getSessionID();
 /*     SynchroNumber currentSynchroNumber = null;
      try {
        currentSynchroNumber = SyncNumberManager.getManager().getSynchroNumber(sessionNumber);
      } catch (SyncNumberNotFoundException ex) {
        LogManager.trace(ex);
        currentSynchroNumber = new SynchroNumber(0,0);
      } */
      String target = header.getTarget();
      String source = header.getSource();
      if (target.indexOf(OpenMSPSyncMessageFactory.OpenMSPTARGET) == -1) {
        res.setStatus(HttpServletResponse.SC_BAD_REQUEST);
      }

      // init client file management
      FileSystemObjectPool pool = FileSystemObjectPoolSelecter.getPool(terminal);
      FileSystem  fileSystem = pool.getFileSystem();

      try {
//        SynchroNumber newSynchroNumber = SyncNumberManager.getManager().getNextSynchroNumber();

        // create answer message
        Message CMLAnswer = MessageFactory.getFactory().createMessage(new Header(sessionNumber+1, target, source , null, null));

        openMLMessage.resetCursor();
        // dispatcher commande ( Get | Map | Results | Sequence | Status | Sync)+, Final?)
        while (openMLMessage.hasMoreMessage()) {
          ContainerMessage CMLRequest = openMLMessage.nextMessage();
          this.dispatchElement(cred, group, CMLRequest, CMLAnswer, terminal, profil, fileSystem);
        }

        // end SyncML code management

        // save CyberML file
        ByteArrayInputStream input = new ByteArrayInputStream(CMLAnswer.encode().getBytes());
        fileSystem.addFile(new MemoryFile("/cyberML.xml", input));
        //generate synchro file
        SynchroFileGenerator fileGenerator = SynchroFileGeneratorManager.getManager().getFileGenerator(terminal);
        fileGenerator.generateSynchroFile(terminal, profil, request, res, fileSystem, Long.toString(sessionNumber+1));

        res.setStatus(HttpServletResponse.SC_OK);
      } finally {
        pool.disposeObject(fileSystem);
      }
    } catch (OpenMSPException ex) {
      res.setStatus(HttpServletResponse.SC_BAD_REQUEST);
      LogManager.trace(new SynchroException(ex));
    } catch (UserNotFoundException ex) {
      res.setStatus(HttpServletResponse.SC_FORBIDDEN);
    } catch(ProfilNotFoundException ex)  {
      LogManager.traceError(0, "OpenMSPService no Profil for user login :"+user.getLogin());
      res.setStatus(HttpServletResponse.SC_FORBIDDEN);
    } catch (Throwable ex) {
      res.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
      LogManager.trace(new SynchroException(ex));
    }
  }

  private final void dispatchElement(Credential cred, String group, ContainerMessage message, Message answer, UserTerminal terminal, Profil profil, FileSystem  fileSystem) {
    Element element = message.getElement();
    try {
      if (element.getElementType() == Element.RESULT)  {
        String serviceName = ((Result)element).getSourceRef();
        String version = null;
        this.callService(cred, serviceName, message, answer, terminal, version, profil, fileSystem);
      } else if (element.getElementType() == Element.STATUS)  {// not supported
        Status status = new  Status (((Status)element).getCmdId(), Status.STATUS_COMMAND_NOT_IMPLEMENTED) ;
        answer.add(status);
      } else if ((element.getElementType() == Element.SYNC)
            || (element.getElementType() == Element.MAP)
            ){
        String serviceName = ((Command)element).getTarget();
        String version = null;
        this.callService(cred, serviceName, message, answer, terminal, version, profil, fileSystem);
      } else if (element.getElementType() == Element.SEQUENCE) {
        while (message.hasMoreMessage())  {
          ContainerMessage commandContainer = message.nextMessage();
          this.dispatchElement(cred, group, commandContainer, answer, terminal, profil, fileSystem);
        }
      } else if (element.getElementType() == Element.FINAL) {
      } else  {
        Status status = new  Status (((RequestCommand)element).getCmdId(), Status.STATUS_WRONG_FORMAT) ;
        answer.add(status);
      }
    } catch (Exception ex)  {
      try {
        LogManager.trace(new SynchroException(ex));
        Status status = new  Status (((RequestCommand)element).getCmdId(), Status.STATUS_FAILED) ;
        answer.add(status);
      } catch (Exception e)  {
        LogManager.trace(new SynchroException(e));
      }
    }
  }

  private final void callService(Credential cred, String serviceName, ContainerMessage message, Message answer, UserTerminal terminal, String version,  Profil profil, FileSystem  fileSystem) throws OpenMSPException {
    try {
      ProxySyncroTarget proxy = SynchroTargerManager.getManager().getProxySynchroTargetForTerminal(terminal);
      if (proxy != null)  {
        proxy.executeOpenMSPCommande(cred, serviceName, message, answer, terminal, version, profil, fileSystem);
      } else  {
        Status status = new  Status (((AbstractCommand)message.getElement()).getCmdId(), Status.STATUS_NOT_FOUND) ;
        answer.add(status);
        LogManager.traceInfo(LogServices.SYNCHROSERVICE, "Service "+serviceName+" ask by CyberML commande and not available");
      }
    } catch (Throwable e)  {
      LogManager.trace(new SynchroException(e));
      Status status = new  Status (((AbstractCommand)message.getElement()).getCmdId(), Status.STATUS_FAILED) ;
      answer.add(status);
    }
  }

  public String getName() {
    return "OpenMSPService";
  }

  public byte[] uncompressAndDecodeRequest(byte[] request) throws IOException {
    // decode compressed data
    byte[] decodedData = GeneralCoder.decodeBase64(request);

    // uncompresse data
    ByteArrayOutputStream byteStream = new ByteArrayOutputStream(1024);
    ByteArrayInputStream inStream = new ByteArrayInputStream(decodedData);
    GZIPInputStream zipStream = new GZIPInputStream(inStream);
    try {
      byte[] buff = new byte[1024];
      int length;
      while ((length = zipStream.read(buff)) != -1) {
        byteStream.write(buff, 0, length);
      }
    } finally  {
      zipStream.close();
      byteStream.close();
    }
    return byteStream.toByteArray();
  }
}